#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Komal Thareja (kthare10@renci.org)
import logging
import os


class Constants:
    DEFAULT_FABRIC_CREDMGR_HOST = "cm.fabric-testbed.net"
    DEFAULT_FABRIC_ORCHESTRATOR_HOST = "orchestrator.fabric-testbed.net"
    DEFAULT_FABRIC_CORE_API_HOST = "uis.fabric-testbed.net"
    DEFAULT_FABRIC_AM_HOST = "artifacts.fabric-testbed.net"
    DEFAULT_FABRIC_BASTION_HOST = "bastion.fabric-testbed.net"
    DEFAULT_LOG_LEVEL = "INFO"
    DEFAULT_LOG_FILE = "/tmp/fablib/fablib.log"
    DEFAULT_DATA_DIR = "/tmp/fablib"
    DEFAULT_WORK_DIR = f"{os.environ['HOME']}/work"
    DEFAULT_FABRIC_CONFIG_DIR = f"{DEFAULT_WORK_DIR}/fabric_config"
    DEFAULT_FABRIC_RC = f"{DEFAULT_FABRIC_CONFIG_DIR}/fabric_rc"
    DEFAULT_TOKEN_LOCATION = f"{os.environ['HOME']}/.tokens.json"
    DEFAULT_SLICE_PRIVATE_KEY_FILE = f"{DEFAULT_FABRIC_CONFIG_DIR}/slice_key"
    DEFAULT_SLICE_PUBLIC_KEY_FILE = f"{DEFAULT_SLICE_PRIVATE_KEY_FILE}.pub"
    DEFAULT_BASTION_KEY_LOCATION = f"{DEFAULT_FABRIC_CONFIG_DIR}/fabric_bastion_key"
    DEFAULT_FABRIC_BASTION_SSH_CONFIG_FILE = f"{DEFAULT_FABRIC_CONFIG_DIR}/ssh_config"
    DEFAULT_FABRIC_SSH_COMMAND_LINE = (
        "ssh -i {{ _self_.private_ssh_key_file }} -F "
        + DEFAULT_FABRIC_BASTION_SSH_CONFIG_FILE
        + " {{ _self_.username }}@{{ _self_.management_ip }}"
    )

    FABRIC_CREDMGR_HOST = "FABRIC_CREDMGR_HOST"
    FABRIC_ORCHESTRATOR_HOST = "FABRIC_ORCHESTRATOR_HOST"
    FABRIC_CORE_API_HOST = "FABRIC_CORE_API_HOST"
    FABRIC_AM_HOST = "FABRIC_AM_HOST"
    FABRIC_TOKEN_LOCATION = "FABRIC_TOKEN_LOCATION"
    FABRIC_PROJECT_ID = "FABRIC_PROJECT_ID"
    FABRIC_BASTION_HOST = "FABRIC_BASTION_HOST"
    FABRIC_BASTION_USERNAME = "FABRIC_BASTION_USERNAME"
    FABRIC_BASTION_KEY_LOCATION = "FABRIC_BASTION_KEY_LOCATION"
    FABRIC_SLICE_PUBLIC_KEY_FILE = "FABRIC_SLICE_PUBLIC_KEY_FILE"
    FABRIC_SLICE_PRIVATE_KEY_FILE = "FABRIC_SLICE_PRIVATE_KEY_FILE"
    FABRIC_SLICE_PRIVATE_KEY_PASSPHRASE = "FABRIC_SLICE_PRIVATE_KEY_PASSPHRASE"
    FABRIC_LOG_FILE = "FABRIC_LOG_FILE"
    FABRIC_LOG_LEVEL = "FABRIC_LOG_LEVEL"
    FABRIC_AVOID = "FABRIC_AVOID"
    FABRIC_SSH_COMMAND_LINE = "FABRIC_SSH_COMMAND_LINE"
    FABLIB_VERSION = "fablib_version"
    FABRIC_BASTION_SSH_CONFIG_FILE = "FABRIC_BASTION_SSH_CONFIG_FILE"

    FABRIC_PRIMARY = "#27aae1"
    FABRIC_PRIMARY_LIGHT = "#cde4ef"
    FABRIC_PRIMARY_DARK = "#078ac1"
    FABRIC_SECONDARY = "#f26522"
    FABRIC_SECONDARY_LIGHT = "#ff8542"
    FABRIC_SECONDARY_DARK = "#d24502"
    FABRIC_BLACK = "#231f20"
    FABRIC_DARK = "#433f40"
    FABRIC_GREY = "#666677"
    FABRIC_LIGHT = "#f3f3f9"
    FABRIC_WHITE = "#ffffff"
    FABRIC_LOGO = "fabric_logo.png"

    FABRIC_PRIMARY_EXTRA_LIGHT = "#dbf3ff"

    SUCCESS_COLOR = "#8eff92"
    SUCCESS_LIGHT_COLOR = "#c3ffc4"
    SUCCESS_DARK_COLOR = "#59cb63"

    ERROR_COLOR = "#ff8589"
    ERROR_LIGHT_COLOR = "#ffb7b9"
    ERROR_DARK_COLOR = "#b34140"

    IN_PROGRESS_COLOR = "#ffff8c"
    IN_PROGRESS_LIGHT_COLOR = "#ffffbe"
    IN_PROGRESS_DARK_COLOR = "#c8555c"

    LOG_LEVELS = {
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
    }

    RUNTIME_SECTION = "runtime"

    CREDMGR_HOST = "credmgr_host"
    ORCHESTRATOR_HOST = "orchestrator_host"
    CORE_API_HOST = "core_api_host"
    AM_HOST = "am_host"
    TOKEN_LOCATION = "token_location"
    PROJECT_ID = "project_id"
    PROJECT_NAME = "project_name"
    BASTION_HOST = "bastion_host"
    BASTION_USERNAME = "bastion_username"
    BASTION_KEY_LOCATION = "bastion_key_location"
    BASTION_KEY_PASSPHRASE = "bastion_key_passphrase"
    LOG_FILE = "log_file"
    LOG_LEVEL = "log_level"
    DATA_DIR = "data_dir"
    SSH_COMMAND_LINE = "ssh_command_line"
    AVOID = "avoid"
    SLICE_PRIVATE_KEY_FILE = "slice_private_key_file"
    SLICE_PUBLIC_KEY_FILE = "slice_public_key_file"
    SLICE_PRIVATE_KEY = "slice_private_key"
    SLICE_PUBLIC_KEY = "slice_public_key"
    SLICE_PRIVATE_KEY_PASSPHRASE = "slice_private_key_passphrase"
    BASTION_SSH_CONFIG_FILE = "bastion_ssh_config_file"

    IMAGE_NAMES = [
        "default_centos8_stream",
        "default_centos9_stream",
        "default_centos_7",
        "default_centos_8",
        "default_debian_11",
        "default_debian_12",
        "default_fedora_39",
        "default_fedora_40",
        "default_freebsd_13_zfs",
        "default_freebsd_14_zfs",
        "default_kali",
        "default_openbsd_7",
        "default_rocky_8",
        "default_rocky_9",
        "default_ubuntu_20",
        "default_ubuntu_22",
        "default_ubuntu_24",
        "docker_rocky_8",
        "docker_rocky_9",
        "docker_ubuntu_20",
        "docker_ubuntu_22",
    ]

    ENV_VAR = "env_var"
    DEFAULT = "default"

    UUID = "uuid"
    NAME = "name"
    BASTION_LOGIN = "bastion_login"
    KEY_TYPE_BASTION = "bastion"
    KEY_TYPE_SLIVER = "sliver"
    PRIVATE_OPENSSH = "private_openssh"
    PUBLIC_OPENSSH = "public_openssh"
    EMAIL = "email"
    SSH_KEYS = "sshkeys"
    EXPIRES_ON = "expires_on"
    LEASE_TIME_FORMAT = "%Y-%m-%d %H:%M:%S %z"
    FABRIC_KEY_TYPE = "fabric_key_type"

    NON_PRETTY_NAME = "non_pretty_name"
    PRETTY_NAME = "pretty_name"
    HEADER_NAME = "header_name"
    AVAILABLE = "Available"
    CAPACITY = "Capacity"
    ALLOCATED = "Allocated"
    VALUE = "value"

    NIC_SHARED_CONNECTX_6 = "SharedNIC-ConnectX-6"
    SMART_NIC_CONNECTX_6 = "SmartNIC-ConnectX-6"
    SMART_NIC_CONNECTX_5 = "SmartNIC-ConnectX-5"
    NVME_P4510 = "NVME-P4510"
    GPU_TESLA_T4 = "GPU-Tesla T4"
    GPU_RTX6000 = "GPU-RTX6000"
    GPU_A30 = "GPU-A30"
    GPU_A40 = "GPU-A40"
    FPGA_XILINX_U280 = "FPGA-Xilinx-U280"
    FPGA_XILINX_SN1022 = "FPGA-Xilinx-SN1022"
    CORES = "Cores"
    RAM = "Ram"
    DISK = "Disk"
    CPUS = "CPUs"
    HOSTS = "Hosts"
    P4_SWITCH = "P4-Switch"

    CMP_NIC_Basic = "NIC_Basic"
    CMP_NIC_ConnectX_6 = "NIC_ConnectX_6"
    CMP_NIC_ConnectX_5 = "NIC_ConnectX_5"
    CMP_NIC_P4 = "NIC_P4"
    CMP_NVME_P4510 = "NVME_P4510"
    CMP_GPU_TeslaT4 = "GPU_TeslaT4"
    CMP_GPU_RTX6000 = "GPU_RTX6000"
    CMP_GPU_A40 = "GPU_A40"
    CMP_GPU_A30 = "GPU_A30"
    CMP_NIC_OpenStack = "NIC_OpenStack"
    CMP_FPGA_Xilinx_U280 = "FPGA_Xilinx_U280"
    CMP_FPGA_Xilinx_SN1022 = "FPGA_Xilinx_SN1022"
    P4_DedicatedPort = "P4_DedicatedPort"

    FABRIC_USER = "fabric"



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)
import os
import time
import traceback
from abc import ABC, abstractmethod
from typing import List

from fabric_cf.orchestrator.orchestrator_proxy import SliceState
from fabrictestbed.slice_editor import (
    Capacities,
    ComponentCatalog,
    ComponentModelType,
    ComponentType,
    ExperimentTopology,
    ServiceType,
)
from fabrictestbed.slice_manager import SliceManager, SliceState, Status
from fabrictestbed.util.constants import Constants


class AbcUtils(ABC):
    bastion_username = os.environ["FABRIC_BASTION_USERNAME"]

    bastion_public_addr = os.environ["FABRIC_BASTION_HOST"]
    bastion_private_ipv4_addr = os.environ["FABRIC_BASTION_HOST_PRIVATE_IPV4"]
    bastion_private_ipv6_addr = os.environ["FABRIC_BASTION_HOST_PRIVATE_IPV6"]
    bastion_key_filename = os.environ["HOME"] + "/.ssh/pruth_fabric_rsa"

    node_ssh_key = None
    with open(os.environ["HOME"] + "/.ssh/id_rsa.pub", "r") as fd:
        node_ssh_key = fd.read().strip()
    node_ssh_key_priv_file = os.environ["HOME"] + "/.ssh/id_rsa"

    @staticmethod
    def create_slice_manager():
        credmgr_host = os.environ[Constants.FABRIC_CREDMGR_HOST]
        orchestrator_host = os.environ[Constants.FABRIC_ORCHESTRATOR_HOST]
        project_id = os.environ[Constants.FABRIC_PROJECT_ID]

        print(f"FABRIC Credential Manager   : {credmgr_host}")
        print(f"FABRIC Orchestrator         : {orchestrator_host}")
        print(f"FABRIC Project Id           : {project_id}")

        slice_manager = SliceManager(
            oc_host=orchestrator_host,
            cm_host=credmgr_host,
            project_id=project_id,
            scope="all",
        )

        # Initialize the slice manager
        slice_manager.initialize()

        return slice_manager

    @abstractmethod
    def run(self, create_slice=True, run_test=True, delete=True):
        """
        Run the test
        :return:
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2024 Illinois Institute of Technology
#
# This software was developed by Illinois Institute of Technology under NSF award 2346499 ("CREASE"),
# as part of the NSF CCRI-CISE (Community Research Infrastructure) program.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Nik Sultana

"""
Methods to work with `attestable switches`.

You would add a switch and operate on it like so::

    from fabrictestbed_extensions.fablib.fablib import FablibManager

    fablib = FablibManager()

    slice = fablib.new_slice(name="MySlice")
    s1 = slice.add_attestable_switch(name="s1", site=site, ports=["p0", "p1"])
    slice.submit();

    s1.start_switch()
    s1.run_command("show_ports")
    #s1.load_program("router.p4")
    s1.run_command("show_tables")
    s1.stop_switch()

    slice.delete()
"""

from __future__ import annotations

import logging
import os
import pickle
import time
from typing import TYPE_CHECKING, Dict, List, Tuple, Union

if TYPE_CHECKING:
    from fabrictestbed_extensions.fablib.slice import Slice

from fabrictestbed.slice_editor import Node as FimNode

from fabrictestbed_extensions.fablib.node import Node


class Attestable_Switch(Node):
    """
    A class that abstracts programmable network elements (switches and NICs). These elements may be attestable -- that is, they provide runtime evidence about their configuration.
    """

    default_cores = 4
    default_ram = 8
    default_disk = 50
    default_image = "attestable_bmv2_v2_ubuntu_20"
    default_username = "ubuntu"
    raw_image = "default_ubuntu_20"
    bmv_prefix = "~/bmv2-remote-attestation/targets/simple_switch/"
    crease_path_prefix = "/home/ubuntu/crease_cfg/"

    __version__ = "beta 2"
    __version_short__ = "b2"

    def __init__(
        self,
        slice: Slice,
        node: FimNode,
        validate: bool = False,
        raise_exception: bool = False,
        ports: List[str] = None,
        from_raw_image=False,
        setup_and_configure=True,
    ):
        """
        Attestable Switch constructor, usually invoked by ``Slice.add_attestable_switch()``.

        :param slice: same meaning as counterpart parameter for `Node`.
        :type slice: Slice

        :param node: same meaning as counterpart parameter for `Node`.
        :type node: Node

        :param validate: same meaning as counterpart parameter for `Node`.
        :type validate: bool

        :param raise_exception: same meaning as counterpart parameter for `Node`.
        :type raise_exception: bool

        :param ports: names of ports that the switch will have.
        :type ports: List[str]

        :param from_raw_image: start from a raw image and install all dependencies -- this takes longer.
        :type from_raw_image: bool

        :param setup_and_configure: set up and configure the attestable switch in post-boot config.
        :type setup_and_configure: bool
        """

        super().__init__(slice, node, validate, raise_exception)

        logging.info(f"Creating Attestable Switch {self.get_name()}.")

        if None == self.get_switch_data(soft=False):
            logging.info(
                f"Attestable Switch {Attestable_Switch.__version__} {self.get_name()}: (not found)"
            )

            assert len(ports) > 0

            self.cfg = {}
            self.cfg["ports"] = ports
            self.cfg["portmap"] = {}
            self.cfg["from_raw_image"] = from_raw_image
            self.cfg["setup_and_configure"] = setup_and_configure

            for port in ports:
                self.cfg["portmap"][port] = (
                    self.add_component(model="NIC_Basic", name=port)
                    .get_interfaces()[0]
                    .get_name()
                )
                logging.info(f"Attestable Switch {self.get_name()}: added port {port}")

            self.set_switch_data(self.cfg)

        else:
            logging.info(
                f"Attestable Switch {Attestable_Switch.__version__} {self.get_name()}: (found)"
            )
            self.cfg = self.get_switch_data(soft=False)

    def get_switch_data(self, soft=True):
        """
        Get switch-specific configuration data.
        """

        if soft:
            return self.cfg
        else:
            if "attestable_switch_config" in self.get_user_data():
                return self.get_user_data()["attestable_switch_config"]
            else:
                return None

    def set_switch_data(self, switch_data: dict):
        """
        Set switch-specific configuration data.
        """

        self.cfg = switch_data
        user_data = self.get_user_data()
        user_data["attestable_switch_config"] = switch_data
        self.set_user_data(user_data)

    def get_switch_config(self, k, quiet=True):
        """
        Get run-time configurable, switch-specific configuration data.
        """

        (out, _) = self.execute(
            f"cat {Attestable_Switch.crease_path_prefix + k}", quiet=quiet
        )
        if out == "None\n":
            return None
        elif out == "True\n":
            return True
        elif out == "False\n":
            return False

    def prep_switch_config_update(self, k, v):
        """
        Support function for commit_switch_config_update().
        """

        return (k, str(v))

    def commit_switch_config_update(self, cfg_update):
        """
        Set run-time configurable, switch-specific configuration data.
        """

        s = ""
        for k, v in cfg_update:
            s += f"echo '{v}' > {Attestable_Switch.crease_path_prefix + k}; "
        self.execute(s)

    def get_port_names(self):
        """
        Get the port names of the switch.
        """

        return self.get_switch_data()["ports"]

    def get_port_interfaces(self):
        """
        Get the interface names of the switch's ports.
        """

        result = {}
        for port in self.get_switch_data()["portmap"].keys():
            result[port] = self.get_slice().get_interface(
                name=self.get_switch_data()["portmap"][port]
            )
        return result

    def get_port_interface(self, port_name):
        """
        Get the interface name of a switch's port name.
        """

        return self.get_slice().get_interface(
            name=self.get_switch_data()["portmap"][port_name]
        )

    def get_port_device_listing(self):
        """
        Get the name-to-interface mapping for a switch.
        """

        mapping = {}
        for ifa in self.get_interfaces():
            mapping[ifa.get_component().get_short_name()] = ifa.get_device_name()
        result = []
        for port in self.get_port_names():
            result.append((port, mapping[port]))
        return result

    def __str__(self):
        """
        Creates a tabulated string describing the properties of the
        node.

        Intended for printing node information.

        :return: Tabulated string of node information
        :rtype: String
        """
        table = [
            ["ID", self.get_reservation_id()],
            ["Name", self.get_name()],
            ["Cores", self.get_cores()],
            ["RAM", self.get_ram()],
            ["Disk", self.get_disk()],
            ["Image", self.get_image()],
            ["Image Type", self.get_image_type()],
            ["Host", self.get_host()],
            ["Site", self.get_site()],
            ["Management IP", self.get_management_ip()],
            ["Reservation State", self.get_reservation_state()],
            ["Error Message", self.get_error_message()],
            ["SSH Command", self.get_ssh_command()],
            ["Ports", str(self.get_port_names())],
        ]

        return tabulate(table)  # , headers=["Property", "Value"])

    @staticmethod
    def name(orig_name):
        """
        Devise a consistent naming prefix for switches.
        """

        assert orig_name is not None
        as_name_prefix = (
            "attestable_switch_" + Attestable_Switch.__version_short__ + "_"
        )
        result = orig_name
        if not orig_name.startswith(as_name_prefix):
            result = as_name_prefix + orig_name
        return result

    def get_name(self) -> str or None:
        """
        Gets the name of the FABRIC node.

        :return: the name of the node
        :rtype: String
        """
        try:
            return Attestable_Switch.name(self.get_fim_node().name)
        except:
            return None

    @staticmethod
    def new_attestable_switch(
        slice: Slice = None,
        name: str = None,
        site: str = None,
        avoid: List[str] = [],
        validate: bool = False,
        raise_exception: bool = False,
        ports: List[str] = None,
        from_raw_image=False,
        setup_and_configure=True,
    ):
        """
        Not intended for API call.  See: Slice.add_attestable_switch()

        Creates a new attestable switch.

        :param slice: the fablib slice to build the new node on
        :type slice: Slice

        :param name: the name of the new node
        :type name: str

        :param site: the name of the site to build the node on
        :type site: str

        :param avoid: a list of node names to avoid
        :type avoid: List[str]

        :param ports: names of ports that the switch will have.
        :type ports: List[str]

        :param from_raw_image: start from a raw image and install all dependencies -- this takes longer.
        :type from_raw_image: bool

        :param setup_and_configure: set up and configure the attestable switch in post-boot config.
        :type setup_and_configure: bool

        :return: a new fablib Attestable_Switch
        :rtype: Attestable_Switch
        """
        if site is None:
            [site] = slice.get_fablib_manager().get_random_sites(avoid=avoid)

        name = Attestable_Switch.name(name)

        logging.info(
            f"Adding attestable switch: {name}, slice: {slice.get_name()}, site: {site}, ports: {ports}, from_raw_image: {from_raw_image}, setup_and_configure: {setup_and_configure}"
        )
        node = Attestable_Switch(
            slice,
            slice.topology.add_node(name=name, site=site),
            validate=validate,
            raise_exception=raise_exception,
            ports=ports,
            from_raw_image=from_raw_image,
            setup_and_configure=setup_and_configure,
        )
        node.set_capacities(
            cores=Node.default_cores, ram=Node.default_ram, disk=Node.default_disk
        )

        node.set_image(
            Attestable_Switch.default_image, username=Attestable_Switch.default_username
        )

        node.init_fablib_data()

        return node

    @staticmethod
    def get_attestable_switch(slice: Slice = None, node=None):
        """
        Returns a fresh reference to a switch using existing FABRIC resources.
        """

        return Attestable_Switch(slice, node)

    @staticmethod
    def get_pretty_name_dict():
        """
        Return mappings from non-pretty names to pretty names.

        Pretty names are in table headers.
        """

        r = Node.get_pretty_name_dict()
        r["ports"] = "Switch ports"
        return r

    def toDict(self, skip=[]):
        """
        Returns the node attributes as a dictionary

        :return: slice attributes as dictionary
        :rtype: dict
        """
        rtn_dict = super().toDict(self, skip)
        if "ports" not in skip:
            rtn_dict["ports"] = str(self.get_port_names())

        return rtn_dict

    def switch_config(self, log_dir="."):
        """
        Post-boot configuration for the switch.
        """

        from_raw_image = self.get_switch_data()["from_raw_image"]

        if self.get_switch_data()["setup_and_configure"]:
            self.execute(
                f"mkdir {Attestable_Switch.crease_path_prefix}; echo 'None' > {Attestable_Switch.crease_path_prefix}with_RA; echo 'None' > {Attestable_Switch.crease_path_prefix}RA_port; echo 'None' > {Attestable_Switch.crease_path_prefix}RA_et; echo 'False' > {Attestable_Switch.crease_path_prefix}Running"
            )

            logging.info(
                f"Attestable Switch {self.get_name()}: starting config. from_raw_image={from_raw_image}"
            )

            if not from_raw_image:
                logging.info(
                    f"Image already contains Attestable Switch: skipping compilation."
                )
            else:
                print(f"Compiling Attestable Switch {self.get_name()}, ", end="")
                start = time.time()
                logging.info(f"Attestable Switch {self.get_name()}: cloning repo...")
                self.execute(
                    'bash -c "git clone https://github.com/awolosewicz/bmv2-remote-attestation.git"',
                    quiet=True,
                )
                self.execute(
                    'bash -c "cd ~/bmv2-remote-attestation && git checkout stable"',
                    quiet=True,
                )
                logging.info(
                    f"Attestable Switch {self.get_name()}: obtaining dependencies..."
                )

                self.execute(
                    "bash -c 'source /etc/lsb-release && echo \"deb http://download.opensuse.org/repositories/home:/p4lang/xUbuntu_${DISTRIB_RELEASE}/ /\" | sudo tee /etc/apt/sources.list.d/home:p4lang.list'",
                    quiet=True,
                )
                self.execute(
                    "bash -c 'source /etc/lsb-release && curl -fsSL https://download.opensuse.org/repositories/home:p4lang/xUbuntu_${DISTRIB_RELEASE}/Release.key | gpg --dearmor | sudo tee /etc/apt/trusted.gpg.d/home_p4lang.gpg > /dev/null'",
                    quiet=True,
                )

                self.execute(
                    'bash -c "sudo apt-get update && sudo apt-get install -y p4lang-p4c net-tools python3-scapy"',
                    quiet=True,
                )
                logging.info(
                    f"Attestable Switch {self.get_name()}: starting compilation..."
                )
                self.execute(
                    'bash -c "cd ~/bmv2-remote-attestation/ && ./install_deps.sh && ./autogen.sh && ./configure && make -j 4"',
                    quiet=True,
                )
                self.execute(
                    'bash -c "sudo mv ~/bmv2-remote-attestation /usr/local/"',
                    quiet=True,
                )
                logging.info(
                    f"Attestable Switch {self.get_name()}: finished compilation"
                )
                print(f"Done! ({time.time() - start:.0f} sec)")

            nothing_p4 = """#include <core.p4>
#include <v1model.p4>

struct metadata {}
struct headers {}

parser DuffParser(packet_in packet,
                  out headers hdr,
                  inout metadata meta,
                  inout standard_metadata_t standard_metadata) {
  state start {
    transition accept;
  }
}

control DuffVerifyChecksum(inout headers hdr,
                           inout metadata meta) {
  apply {}
}

control DuffIngress(inout headers hdr,
                    inout metadata meta,
                    inout standard_metadata_t standard_metadata) {
  apply {}
}

control DuffEgress(inout headers hdr,
                   inout metadata meta,
                   inout standard_metadata_t standard_metadata) {
  apply {}
}

control DuffComputeChecksum(inout headers  hdr,
                            inout metadata meta) {
  apply {}
}

control DuffDeparser(packet_out packet, in headers hdr) {
  apply {}
}

V1Switch(
  DuffParser(),
  DuffVerifyChecksum(),
  DuffIngress(),
  DuffEgress(),
  DuffComputeChecksum(),
  DuffDeparser()
) main;"""

            print(f"Setting up Attestable Switch {self.get_name()}, ", end="")
            start = time.time()

            self.execute("sudo sysctl net.ipv4.ip_forward=1", quiet=True)

            self.execute(
                f"echo '{nothing_p4}' > {Attestable_Switch.crease_path_prefix}nothing.p4"
            )

            for port in self.get_port_names():
                for (
                    ifa
                ) in (
                    self.get_interfaces()
                ):  # FIXME inefficient code -- use a look-up instead of looping.
                    if ifa.get_component().get_short_name() == port:
                        self.execute(
                            f"sudo ip link set dev {ifa.get_device_name()} up",
                            quiet=True,
                        )
                        self.execute(
                            f"sudo ip link set dev {ifa.get_device_name()} arp off"
                        )

            # self.execute(f"sudo ip route del 0/0")
            logging.info(f"Attestable Switch {self.get_name()}: finished config")
            print(f"Done! ({time.time() - start:.0f} sec)")
        else:
            logging.info(
                f"Attestable Switch {self.get_name()}: skipping setup and config"
            )

    def check(self, quiet=True):
        """
        Check whether the switch was initialised correctly.
        """

        result = True

        (out, _) = self.execute("sudo sysctl net.ipv4.ip_forward", quiet=True)
        check = out == "net.ipv4.ip_forward = 1\n"
        if not check:
            logging.error(f"Attestable Switch {self.get_name()}: failed check 1")
            if not quiet:
                print(f"Attestable Switch {self.get_name()}: failed check 1")

        result = result and check

        (out, _) = self.execute(
            f"ls -s {Attestable_Switch.crease_path_prefix}nothing.p4", quiet=True
        )
        check = out == f"4 {Attestable_Switch.crease_path_prefix}nothing.p4\n"
        if not check:
            logging.error(f"Attestable Switch {self.get_name()}: failed check 2")
            if not quiet:
                print(f"Attestable Switch {self.get_name()}: failed check 2")

        result = result and check

        return result

    def start_switch(
        self,
        program="/home/ubuntu/crease_cfg/nothing.json",
        dry=False,
        quiet=True,
        force=False,
        with_RA=False,
        RA_port=None,
        RA_et=None,
    ):
        """
        Start the switch executing, and have it run a P4 program.
        """

        if not force and self.get_switch_config("Running"):
            print("Switch already running")
            return False

        port_sequence = ""
        port_num = 0
        for _, port_device in self.get_port_device_listing():
            port_sequence += f"-i {str(port_num)}@{port_device} "
            port_num += 1

        cfg_update = []

        RA_inclusion = ""
        cfg_update.append(self.prep_switch_config_update("with_RA", False))

        if with_RA:
            RA_inclusion = "--enable-ra"
            cfg_update.append(self.prep_switch_config_update("with_RA", True))

        if None != RA_port:
            assert with_RA
            cfg_update.append(self.prep_switch_config_update("RA_port", RA_port))
            RA_inclusion += " --ra-port " + str(RA_port)
        else:
            cfg_update.append(self.prep_switch_config_update("RA_port", None))

        if None != RA_et:
            assert with_RA
            cfg_update.append(self.prep_switch_config_update("RA_et", RA_et))
            RA_inclusion += " --ra-etype " + str(RA_et)
        else:
            cfg_update.append(self.prep_switch_config_update("RA_et", None))

        commands = [
            "[ ! -d ~/bmv2-remote-attestation ] && cd ~ && sudo ln -s /usr/local/bmv2-remote-attestation",
            f"[ ! -f {Attestable_Switch.crease_path_prefix}nothing.json ] && cd {Attestable_Switch.crease_path_prefix} && p4c --target bmv2 --arch v1model {Attestable_Switch.crease_path_prefix}nothing.p4",
            f'nohup bash -c "sudo {Attestable_Switch.bmv_prefix}simple_switch {port_sequence} {program} --log-file ~/switch.log --log-flush -- --enable-swap {RA_inclusion}" &',
        ]

        stdout = []
        stderr = []
        if dry:
            for command in commands:
                print(command)
        else:
            for command in commands:
                (out, err) = self.execute(command, quiet=quiet)
                stdout.append(out)
                stderr.append(err)

        stderr = list(filter(lambda line: line != "", stderr))

        if not quiet:
            print("stderr: " + str(stderr))

        result = None

        if [] == stderr:
            cfg_update.append(self.prep_switch_config_update("Running", True))
            result = True
        else:
            result = False

        self.commit_switch_config_update(cfg_update)

        return result

    def stop_switch(self, dry=False, quiet=True, force=False):
        """
        Stop the switch from executing.
        """

        if not force and not self.get_switch_config("Running"):
            print("Switch not running")
            return False

        command = "sudo killall simple_switch"

        cfg_update = []

        stdout = []
        stderr = []
        if dry:
            print(command)
        else:
            (out, err) = self.execute(command, quiet=quiet)
            stdout.append(out)
            stderr.append(err)

        stderr = list(filter(lambda line: line != "", stderr))

        if not quiet:
            print("stderr: " + str(stderr))

        result = None

        if force or [] == stderr:
            cfg_update.append(self.prep_switch_config_update("Running", False))
            result = True
        else:
            result = False

        self.commit_switch_config_update(cfg_update)

        return result

    def load_program(self, filename, dry=False, quiet=True):
        """
        Runtime update of a P4 program on a running switch.
        """

        output_file = os.path.splitext(os.path.basename(filename))[0] + ".json"
        commands = [
            f"p4c --target bmv2 --arch v1model ~/{os.path.basename(filename)}",
            f'echo "load_new_config_file {output_file}" | {Attestable_Switch.bmv_prefix}simple_switch_CLI',
            f'echo "swap_configs" | {Attestable_Switch.bmv_prefix}simple_switch_CLI',
        ]

        stdout = []
        stderr = []
        if dry:
            for command in commands:
                print(command)
        else:
            self.upload_file(filename, os.path.basename(filename), retry=1)
            for command in commands:
                (out, err) = self.execute(command, quiet=quiet)
                stdout.append(out)
                stderr.append(err)

        stderr = list(filter(lambda line: line != "", stderr))

        if [] == stderr:
            return True
        else:
            return False

    def run_command(self, cmd, dry=False, quiet=False):
        """
        Run a CLI command on the switch.
        """

        command = f"echo '{cmd}' | {Attestable_Switch.bmv_prefix}simple_switch_CLI"

        stdout = []
        stderr = []

        if dry:
            print(command)
        else:
            (out, err) = self.execute(command, quiet=quiet)
            stdout.append(out)
            stderr.append(err)

        for out in stdout:
            for line in out.split("\n"):
                if "RuntimeCmd: Error" in line:
                    return False

        stderr = list(filter(lambda line: line != "", stderr))

        if [] == stderr:
            return True
        else:
            return False

    def get_switch_features(self):
        """
        Get feature information from the switch.
        """

        result = {}
        result["Running"] = self.get_switch_config("Running")
        if self.get_switch_config("Running"):
            result["with_RA"] = self.get_switch_config("with_RA")
            if self.get_switch_config("with_RA"):
                if None != self.get_switch_config("RA_port"):
                    result["RA_port"] = self.get_switch_config("RA_port")
                if None != self.get_switch_config("RA_et"):
                    result["RA_et"] = self.get_switch_config("RA_et")
        return result

    def get_version(self):
        """
        Get version information from the switch.
        """

        commands = [
            "[ ! -d ~/bmv2-remote-attestation ] && cd ~ && sudo ln -s /usr/local/bmv2-remote-attestation",
            f"{Attestable_Switch.bmv_prefix}simple_switch -v",
        ]

        stdout = []
        stderr = []

        for command in commands:
            (out, err) = self.execute(command, quiet=False)
            stdout.append(out)
            stderr.append(err)

        stderr = list(filter(lambda line: line != "", stderr))

        if [] == stderr:
            print(str(self.get_switch_features()))
            return True
        else:
            return False



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Komal Thareja(kthare10@renci.org)

import os
import tarfile

import ipywidgets as widgets
from IPython.display import display

from fabrictestbed_extensions.fablib.constants import Constants
from fabrictestbed_extensions.fablib.fablib import FablibManager


class ArtifactManagerUI:
    """
    A UI class for managing and interacting with artifacts using the FablibManager.

    Attributes
    ----------
    download_dir : str
        The directory where artifacts will be downloaded.
    fablib : FablibManager
        Instance of FablibManager for interacting with artifacts.
    artifacts : list
        List of artifacts retrieved from FablibManager.
    error_output : widgets.Output
        Widget to display error messages.
    title_filter : widgets.Text
        Text input widget for filtering artifacts by title.
    tag_filter : widgets.Text
        Text input widget for filtering artifacts by tags.
    project_filter : widgets.Text
        Text input widget for filtering artifacts by project.
    grid : widgets.GridBox
        GridBox widget displaying the artifact information.
    """

    def __init__(self, fablib: FablibManager, download_dir: str = "/home/fabric/work"):
        """
        Initializes the ArtifactManagerUI with a specified download directory.

        Parameters
        ----------
        download_dir : str, optional
            The directory where artifacts will be downloaded (default is "/home/fabric/work").
        """
        self.download_dir = download_dir
        self.fablib = fablib
        self.artifacts = self.fablib.get_manager().list_artifacts()
        self.error_output = widgets.Output()

        # Initialize title and tag filters
        self.title_filter = widgets.Text(
            description="Title:",
            placeholder="Enter title keyword",
            layout=widgets.Layout(width="50%"),
        )
        self.tag_filter = widgets.Text(
            description="Tag:",
            placeholder="Enter tag keyword",
            layout=widgets.Layout(width="50%"),
        )

        self.title_filter.observe(self.filter_artifacts, names="value")
        self.tag_filter.observe(self.filter_artifacts, names="value")

        # Initialize grid as None
        self.grid = None

    def download_artifact_handler(
        self, artifact_title: str, version_urn: str, version: str
    ):
        """
        Handler for downloading and extracting artifacts.

        Parameters
        ----------
        artifact_title : str
            The title of the artifact to download.
        version_urn : str
            The URN of the version to download.
        version : str
            The version of the artifact to download.

        Raises
        ------
        Exception
            If there is an error during the download or extraction process.
        """
        with self.error_output:
            try:
                self.error_output.clear_output()  # Clear any previous error messages
                location = self.fablib.download_artifact(
                    download_dir=self.download_dir,
                    version_urn=version_urn,
                    version=version,
                )
                print(
                    f"Artifact '{artifact_title}' version: '{version}' downloaded to {location} successfully!"
                )

                # Extract the tar file at the same location
                if tarfile.is_tarfile(location):
                    with tarfile.open(location, "r:*") as tar:
                        extract_path = os.path.dirname(location)
                        tar.extractall(path=extract_path)
                        print(
                            f"Artifact '{artifact_title}' version: '{version}' extracted to {extract_path} successfully!"
                        )
                else:
                    print(f"Downloaded file '{location}' is not a valid tar file.")

            except Exception as e:
                print(f"Failed to download artifact: {e}")

    def filter_artifacts(self, change):
        """
        Filters artifacts based on the title, tag or project filter text input.

        Parameters
        ----------
        change : dict
            The change dictionary containing the new value of the text input.
        """
        filter_text = self.title_filter.value.lower()
        filtered_artifacts = [
            artifact
            for artifact in self.artifacts
            if filter_text in artifact["title"].lower()
        ]
        tag_text = self.tag_filter.value.lower()
        filtered_artifacts = [
            artifact
            for artifact in filtered_artifacts
            if tag_text in ", ".join(artifact["tags"])
        ]

        self.update_ui(filtered_artifacts)

    def update_ui(self, artifacts):
        """
        Updates the UI with the given list of artifacts.

        Parameters
        ----------
        artifacts : list
            The list of artifacts to be displayed in the UI.
        """
        # Clear existing grid content, if any
        if self.grid is not None:
            self.grid.children = (
                []
            )  # Clear the grid's content without removing the widget itself

        # Define table headers
        headers = [
            widgets.HTML(value="<b>Title</b>"),
            widgets.HTML(value="<b>Description</b>"),
            widgets.HTML(value="<b>Tags</b>"),
            widgets.HTML(value="<b>Versions</b>"),
            widgets.HTML(value="<b>Download</b>"),
            widgets.HTML(value="<b>Project</b>"),
            widgets.HTML(value="<b>Authors</b>"),
        ]

        for h in headers:
            h.style.text_align = "center"
            h.style.text_color = Constants.FABRIC_WHITE
            h.style.background = Constants.FABRIC_PRIMARY

        # Prepare grid items
        grid_items = headers.copy()

        # Build interactive UI for each artifact
        for artifact in artifacts:
            title_label = widgets.HTML(value=artifact["title"])
            authors_label = widgets.HTML(
                value=", ".join([author["name"] for author in artifact["authors"]])
            )
            description_label = widgets.HTML(value=artifact["description_short"])
            tags_label = widgets.HTML(value=", ".join(artifact["tags"]))
            project_label = widgets.HTML(value=artifact.get("project_name", "N/A"))

            # Version dropdown list
            version_options = [
                (f"v{version['version']}", version["urn"])
                for version in artifact["versions"]
            ]
            version_dropdown = widgets.Dropdown(
                options=version_options,
                description="",
                layout=widgets.Layout(width="auto"),
            )

            # Download button
            download_button = widgets.Button(
                description="Download",
                tooltip="Download selected version",
                layout=widgets.Layout(width="auto"),
            )
            download_button.style.button_color = Constants.FABRIC_PRIMARY_LIGHT

            # Bind the download handler
            download_button.on_click(
                lambda b, artifact_title=artifact[
                    "title"
                ], version_dropdown=version_dropdown: self.download_artifact_handler(
                    artifact_title, version_dropdown.value, version_dropdown.label
                )
            )

            # Append artifact details and controls to grid items
            grid_items.extend(
                [
                    title_label,
                    description_label,
                    tags_label,
                    version_dropdown,
                    download_button,
                    project_label,
                    authors_label,
                ]
            )

        # Create and display the grid layout
        if self.grid is None:
            self.grid = widgets.GridBox(
                grid_items,
                layout=widgets.Layout(
                    grid_template_columns="20% 30% 10% 10% 10% 10% 10%",
                    grid_gap="10px",
                    align_items="center",
                ),
            )
        else:
            self.grid.children = grid_items

        # Display the filter and the grid
        display(widgets.HBox([self.title_filter, self.tag_filter]))
        display(self.grid)
        display(self.error_output)

    def create_ui(self):
        """
        Creates and displays the user interface for interacting with artifacts.
        """
        self.update_ui(self.artifacts)


if __name__ == "__main__":
    # Create an instance of the class and render the UI
    artifact_manager_ui = ArtifactManagerUI(fablib=FablibManager())
    artifact_manager_ui.create_ui()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following nditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

"""
Methods to work with FABRIC network interfaces.
"""

from __future__ import annotations

import ipaddress
import json
import logging
from ipaddress import IPv4Address
from typing import TYPE_CHECKING, Any, List, Union

import jinja2
from fabrictestbed.slice_editor import Flags
from fim.user import Capacities, InterfaceType, Labels
from tabulate import tabulate

from fabrictestbed_extensions.fablib.constants import Constants

if TYPE_CHECKING:
    from fabrictestbed_extensions.fablib.slice import Slice
    from fabrictestbed_extensions.fablib.node import Node
    from fabrictestbed_extensions.fablib.network_service import NetworkService
    from fabrictestbed_extensions.fablib.component import Component
    from fabrictestbed_extensions.fablib.facility_port import FacilityPort
    from fabrictestbed_extensions.fablib.switch import Switch

from fabrictestbed.slice_editor import UserData
from fim.user.interface import Interface as FimInterface


class Interface:
    CONFIGURED = "configured"
    MODE = "mode"
    AUTO = "auto"
    MANUAL = "manual"
    ADDR = "addr"
    CONFIG = "config"

    def __init__(
        self,
        component: Component = None,
        fim_interface: FimInterface = None,
        node: Union[Switch, FacilityPort] = None,
        model: str = None,
        parent: Interface = None,
    ):
        """
        .. note::

            Objects of this class are not created directly.

        :param component: the component to set on this interface
        :type component: Component

        :param fim_interface: the FABRIC information model interface
            to set on this fablib interface
        :type fim_interface: FimInterface

        :param node: the facility Port to which interface is assoicated with
        :type node: FacilityPort
        """
        super().__init__()
        self.fim_interface = fim_interface
        self.component = component
        self.network = None
        self.dev = None
        self.node = node
        self.model = model
        self.interfaces = None
        self.parent = parent

    def get_fablib_manager(self):
        """
        Get a reference to :py:class:`.FablibManager` instance.
        """
        return self.get_slice().get_fablib_manager()

    def __str__(self):
        """
        Creates a tabulated string describing the properties of the interface.

        Intended for printing interface information.

        :return: Tabulated string of interface information
        :rtype: String
        """
        if self.get_network():
            network_name = self.get_network().get_name()
        else:
            network_name = None

        table = [
            ["Name", self.get_name()],
            ["Network", network_name],
            ["Bandwidth", self.get_bandwidth()],
            ["Mode", self.get_mode()],
            ["VLAN", self.get_vlan()],
            ["MAC", self.get_mac()],
            ["Physical Device", self.get_physical_os_interface_name()],
            ["Device", self.get_device_name()],
            ["Address", self.get_ip_addr()],
            ["Numa Node", self.get_numa_node()],
            ["Switch Port", self.get_switch_port()],
        ]

        subnet = self.get_subnet()
        if subnet:
            table.append(["Subnet", subnet])
        peer_subnet = self.get_peer_subnet()
        if peer_subnet:
            table.append(["Peer Subnet", peer_subnet])

        peer_asn = self.get_peer_asn()
        if peer_asn:
            table.append(["Peer ASN", peer_asn])

        peer_bgp = self.get_peer_bgp_key()
        if peer_bgp:
            table.append(["Peer BGP Key", peer_bgp])

        peer_account_id = self.get_peer_account_id()
        if peer_account_id:
            table.append(["Peer Account Id", peer_account_id])

        return tabulate(table)

    def toJson(self):
        """
        Returns the interface attributes as a json string

        :return: slice attributes as json string
        :rtype: str
        """
        return json.dumps(self.toDict(), indent=4)

    @staticmethod
    def get_pretty_name_dict():
        """
        Return a mapping used when rendering table headers.
        """
        return {
            "name": "Name",
            "short_name": "Short Name",
            "node": "Node",
            "network": "Network",
            "bandwidth": "Bandwidth",
            "vlan": "VLAN",
            "mac": "MAC",
            "physical_dev": "Physical Device",
            "dev": "Device",
            "username": "Username",
            "management_ip": "Management IP",
            "state": "State",
            "error": "Error",
            "ssh_command": "SSH Command",
            "public_ssh_key_file": "Public SSH Key File",
            "private_ssh_key_file": "Private SSH Key File",
            "mode": "Mode",
            "ip_addr": "IP Address",
            "numa": "Numa Node",
            "switch_port": "Switch Port",
        }

    def toDict(self, skip=[]):
        """
        Returns the interface attributes as a dictionary

        :return: slice attributes as dictionary
        :rtype: dict
        """
        if self.get_network():
            logging.info(
                f"Getting results from get network name thread for iface {self.get_name()} "
            )
            network_name = self.get_network().get_name()
        else:
            network_name = None

        if self.get_node():
            logging.info(
                f"Getting results from get node name thread for iface {self.get_name()} "
            )
            node_name = self.get_node().get_name()
        else:
            node_name = None

        if self.get_node() and str(self.get_node().get_reservation_state()) == "Active":
            mac = str(self.get_mac())
            physical_dev = str(self.get_physical_os_interface_name())
            dev = str(self.get_device_name())
            ip_addr = str(self.get_ip_addr())
        else:
            mac = ""
            physical_dev = ""
            dev = ""
            ip_addr = ""

        return {
            "name": str(self.get_name()),
            "short_name": str(self.get_short_name()),
            "node": str(node_name),
            "network": str(network_name),
            "bandwidth": str(self.get_bandwidth()),
            "mode": str(self.get_mode()),
            "vlan": (
                str(self.get_vlan()) if self.get_vlan() else ""
            ),  # str(self.get_vlan()),
            "mac": mac,
            "physical_dev": physical_dev,
            "dev": dev,
            "ip_addr": ip_addr,
            "numa": str(self.get_numa_node()),
            "switch_port": str(self.get_switch_port()),
        }

    def get_switch_port(self) -> str:
        """
        Get the name of the port on the switch corresponding to this interface

        :return: name of the port on switch
        :rtype: String
        """
        network = self.get_network()
        if network and network.get_fim():
            ifs = None
            for ifs_name in network.get_fim().interfaces.keys():
                if self.get_name() in ifs_name:
                    ifs = network.get_fim().interfaces[ifs_name]
                    break
            if ifs and ifs.labels and ifs.labels.local_name:
                return ifs.labels.local_name

    def get_numa_node(self) -> str:
        """
        Retrieve the NUMA node of the component linked to the interface.

        :return: NUMA node of the linked component.
        :rtype: str
        """
        if self.get_component() is not None:
            return self.get_component().get_numa_node()

    def generate_template_context(self):
        context = self.toDict()
        return context

    def get_template_context(self):
        return self.get_slice().get_template_context(self)

    def render_template(self, input_string):
        environment = jinja2.Environment()
        # environment.json_encoder = json.JSONEncoder(ensure_ascii=False)

        template = environment.from_string(input_string)
        output_string = template.render(self.get_template_context())

        return output_string

    def show(
        self,
        fields=None,
        output: str = None,
        quiet: bool = False,
        colors: bool = False,
        pretty_names: bool = True,
    ):
        """
        Show a table containing the current interface attributes.

        There are several output options: "text", "pandas", and "json" that determine the format of the
        output that is returned and (optionally) displayed/printed.

        output:  'text': string formatted with tabular
                  'pandas': pandas dataframe
                  'json': string in json format

        fields: json output will include all available fields.

        Example: fields=['Name','MAC']

        :param output: output format
        :type output: str
        :param fields: list of fields to show
        :type fields: List[str]
        :param quiet: True to specify printing/display
        :type quiet: bool
        :param colors: True to specify state colors for pandas output
        :type colors: bool
        :param pretty_names: Display pretty names
        :type pretty_names: bool
        :return: table in format specified by output parameter
        :rtype: Object
        """

        data = self.toDict()

        if pretty_names:
            pretty_names_dict = self.get_pretty_name_dict()
        else:
            pretty_names_dict = {}

        table = self.get_fablib_manager().show_table(
            data,
            fields=fields,
            title="Interface",
            output=output,
            quiet=quiet,
            pretty_names_dict=pretty_names_dict,
        )

        return table

    def set_auto_config(self):
        """
        Enable auto-configuration for the interface.

        This method sets the `auto_config` flag to `True` for the interface
        associated with the current instance. The `auto_config` flag enables
        automatic configuration of the interface by Control Framework.

        :return: None
        """
        fim_iface = self.get_fim()
        fim_iface.flags = Flags(auto_config=True)

    def unset_auto_config(self):
        """
        Disable auto-configuration for the interface.

        This method sets the `auto_config` flag to `False` for the interface
        associated with the current instance. The `auto_config` flag disables
        automatic configuration of the interface by Control Framework.

        :return: None
        """
        fim_iface = self.get_fim()
        fim_iface.flags = Flags(auto_config=False)

    def get_peer_port_name(self) -> str or None:
        """
        If available provide the name of the attached port on the dataplane switch.
        Only possible once the slice has been instantiated.
        """
        if (
            self.fim_interface
            and self.fim_interface.get_peers()
            and self.fim_interface.get_peers()[0]
        ):
            return self.fim_interface.get_peers()[0].labels.local_name
        else:
            return None

    def get_peer_port_vlan(self) -> str:
        """
        Returns the VLAN associated with the interface.
        For shared NICs extracts it from label_allocations.

        :return: VLAN to be used for Port Mirroring
        :rtype: String
        """
        vlan = self.get_vlan()
        if not vlan:
            label_allocations = self.get_fim().get_property(pname="label_allocations")
            if label_allocations:
                return label_allocations.vlan

    def get_device_name(self) -> str:
        """
        Gets a name of the device name on the node

        If the interface requires a FABRIC VLAN tag, the interface name returned
        will be the VLAN tagged interface name.

        :return: OS interface name
        :rtype: String
        """
        try:
            fablib_data = self.get_fablib_data()
            if "dev" in fablib_data and fablib_data.get("dev"):
                return fablib_data.get("dev")
            else:
                # logging.debug(f"iface: {self}")
                os_iface = self.get_physical_os_interface_name()
                vlan = self.get_vlan()

                fablib_data["base_dev"] = os_iface

                if os_iface and vlan:
                    os_iface = f"{os_iface}.{vlan}"

                fablib_data["dev"] = os_iface

                if os_iface:
                    self.set_fablib_data(fablib_data)
            return os_iface

        except Exception as e:
            logging.error(f"get_device_name: error occurred - e: {e}")

    def get_os_interface(self) -> str:
        """
        Gets a name of the interface the operating system uses for this
        FABLib interface.

        If the interface requires a FABRIC VLAN tag, the interface name retruned
        will be the VLAN tagged.

        :return: OS interface name
        :rtype: String

        .. deprecated:: 1.6.5
           Use `get_device_name()` instead.
        """
        try:
            # logging.debug(f"iface: {self}")
            os_iface = self.get_physical_os_interface_name()
            vlan = self.get_vlan()

            if vlan is not None:
                os_iface = f"{os_iface}.{vlan}"
        except:
            os_iface = None

        return os_iface

    def get_mac(self) -> str:
        """
        Gets the MAC address of the interface.

        :return: MAC address
        :rtype: String
        """
        try:
            if self.parent:
                mac = self.parent.get_mac()
            else:
                mac = self.get_fim().get_property(pname="label_allocations").mac
        except:
            mac = None

        return mac

    def get_os_dev(self):
        """
        Gets json output of 'ip addr list' for the interface.

        :return: device description
        :rtype: Dict
        """
        if not self.dev:
            ip_addr_list_json = self.get_node().ip_addr_list(output="json")

            mac = self.get_mac()
            for dev in ip_addr_list_json:
                if str(dev["address"].upper()) == str(mac.upper()):
                    self.dev = dev
                    return dev
        else:
            return self.dev

        return None

    def get_physical_os_interface_name(self) -> str:
        """
        Gets a name of the physical interface the operating system uses for this
        FABLib interface.

        If the interface requires a FABRIC VLAN tag, the base interface name
        will be returned (i.e. not the VLAN tagged interface)

        :return: physical OS interface name
        :rtype: String
        """
        try:
            return self.get_os_dev()["ifname"]
        except:
            return None

    def config_vlan_iface(self):
        """
        Configure vlan interface

        NOTE: Not intended for API use
        """
        if self.get_vlan() is not None:
            self.get_node().add_vlan_os_interface(
                os_iface=self.get_physical_os_interface_name(),
                vlan=self.get_vlan(),
                interface=self,
            )

    def set_ip(self, ip=None, cidr=None, mtu=None):
        """
        Configures IP address for the interface inside the VM

        :param ip: IP address
        :type ip: String

        :param cidr: CIDR
        :type cidr: String

        :param mtu: MTU
        :type mtu: String

        .. deprecated:: 1.7.0
           Use `set_ip_os_interface()` instead.
        """
        if cidr:
            cidr = str(cidr)
        if mtu:
            mtu = str(mtu)

        self.get_node().set_ip_os_interface(
            os_iface=self.get_physical_os_interface_name(),
            vlan=self.get_vlan(),
            ip=ip,
            cidr=cidr,
            mtu=mtu,
        )

    def ip_addr_add(self, addr, subnet):
        """
        Add an IP address to the interface in the node.

        :param addr: IP address
        :type addr: IPv4Address or IPv6Address
        :param subnet: subnet
        :type subnet: IPv4Network or IPv4Network
        """
        self.get_node().ip_addr_add(addr, subnet, self)

    def ip_addr_del(self, addr, subnet):
        """
        Delete an IP address to the interface in the node.

        :param addr: IP address
        :type addr: IPv4Address or IPv6Address
        :param subnet: subnet
        :type subnet: IPv4Network or IPv4Network
        """
        self.get_node().ip_addr_del(addr, subnet, self)

    def ip_link_up(self):
        """
        Bring up the link on the interface.
        """
        if self.get_network():
            self.get_node().ip_link_up(None, self)

    def ip_link_down(self):
        """
        Bring down the link on the interface.

        """
        self.get_node().ip_link_down(None, self)

    def ip_link_toggle(self):
        """
        Toggle the dev down then up.

        """
        self.get_node().ip_link_down(None, self)
        self.get_node().ip_link_up(None, self)

    def un_manage_interface(self):
        """
        Mark an interface unmanaged by Network Manager;
        This is needed to be run on rocky* images to avoid the
        network configuration from being overwritten by NetworkManager
        """
        if self.get_network() is None:
            return

        self.get_node().un_manage_interface(self)

    def set_vlan(self, vlan: Any = None):
        """
        Set the VLAN on the FABRIC request.

        :param vlan: vlan
        :type vlan: String or int
        """
        if vlan:
            vlan = str(vlan)

        if_labels = self.get_fim().get_property(pname="labels")
        if_labels.vlan = str(vlan)
        self.get_fim().set_properties(labels=if_labels)

        return self

    def set_bandwidth(self, bw: int):
        """
        Set the Bandwidths on the FABRIC request.

        :param addr: bw
        :type addr: int
        """
        if not bw:
            return

        if_capacities = self.get_fim().get_property(pname="capacities")
        if_capacities.bw = int(bw)
        self.get_fim().set_properties(capacities=if_capacities)

        return self

    def get_fim_interface(self) -> FimInterface:
        """
        Not recommended for most users.

        Gets the node's FABRIC Information Model (fim) object. This method
        is used to access data at a lower level than FABlib.

        :return: the FABRIC model node
        :rtype: fim interface

        .. deprecated:: 1.7.0
           Use `get_fim()` instead.
        """
        return self.get_fim()

    def get_bandwidth(self) -> int:
        """
        Gets the bandwidth of an interface. Basic NICs claim 0 bandwidth but
        are 100 Gbps shared by all Basic NICs on the host.

        :return: bandwith
        :rtype: String
        """
        if self.get_component() and self.get_component().get_model() == "NIC_Basic":
            return 100
        elif self.get_fim() and self.get_fim().capacities:
            return self.get_fim().capacities.bw

    def get_vlan(self) -> str:
        """
        Gets the FABRIC VLAN of an interface.

        :return: VLAN
        :rtype: String
        """
        try:
            vlan = self.get_fim().get_property(pname="labels").vlan
        except:
            vlan = None
        return vlan

    def get_reservation_id(self) -> str or None:
        """
        Gets the reservation id

        :return: reservation id
        :rtype: String
        """
        try:
            return self.get_fim().get_property(pname="reservation_info").reservation_id
        except:
            return None

    def get_reservation_state(self) -> str or None:
        """
        Gets the reservation state

        :return: reservation state
        :rtype: String
        """
        try:
            return (
                self.get_fim().get_property(pname="reservation_info").reservation_state
            )
        except:
            return None

    def get_error_message(self) -> str:
        """
        Gets the error messages

        :return: error
        :rtype: String
        """
        try:
            return self.get_fim().get_property(pname="reservation_info").error_message
        except:
            return ""

    def get_short_name(self):
        """
        Retrieve the shortened name of the interface.

        This method strips off the extra parts of the name added by the FIM. Specifically, it removes the
        prefix formed by concatenating the node name and the component's short name
        followed by a hyphen.

        :return: Shortened name of the interface.
        :rtype: str
        """
        if self.parent:
            return self.get_name()

        # Strip off the extra parts of the name added by FIM
        prefix_length = len(
            f"{self.get_node().get_name()}-{self.get_component().get_short_name()}-"
        )
        return self.get_name()[prefix_length:]

    def get_name(self) -> str:
        """
        Gets the name of this interface.

        :return: the name of this interface
        :rtype: String
        """
        return self.get_fim().name

    def get_component(self) -> Component:
        """
        Gets the component attached to this interface.

        :return: the component on this interface
        :rtype: Component
        """
        return self.component

    def get_model(self) -> str:
        """
        Gets the component model type on this interface's component.

        :return: the model of this interface's component
        :rtype: str
        """
        try:
            if self.model:
                return self.model
            elif self.node:
                return self.node.get_model()
            else:
                return self.get_component().get_model()
        except Exception:
            return ""

    def get_site(self) -> str:
        """
        Gets the site this interface's component is on.

        :return: the site this interface is on
        :rtype: str
        """
        return self.get_node().get_site()

    def get_slice(self) -> Slice:
        """
        Gets the FABLIB slice this interface's node is attached to.

        :return: the slice this interface is attached to
        :rtype: Slice
        """
        return self.get_node().get_slice()

    def get_node(self) -> Union[Node, FacilityPort]:
        """
        Gets the node this interface's component is on.

        :return: the node this interface is attached to
        :rtype: Node
        """
        if self.node:
            return self.node
        else:
            return self.get_component().get_node()

    def get_network(self) -> NetworkService:
        """
        Gets the network this interface is on.

        :return: the network service this interface is on
        :rtype: NetworkService
        """

        if self.network is not None:
            logging.debug(
                f"Interface known network. Returning known network for interface {self.get_name()}"
            )
            return self.network
        else:
            logging.debug(
                f"Interface does not known network. Finding network for interface {self.get_name()}"
            )

            for net in self.get_slice().get_networks():
                if net.has_interface(self):
                    self.network = net
                    logging.debug(
                        f"Interface network found. interface {self.get_name()}, network {self.network.get_name()}"
                    )
                    return self.network

    # fablib.Interface.get_ip_link()
    def get_ip_link(self):
        """
        Gets the ip link info for this interface.

        :return ip link info
        :rtype: str
        """
        try:
            stdout, stderr = self.get_node().execute("ip -j link list", quiet=True)

            links = json.loads(stdout)

            dev = self.get_device_name()
            if dev is None:
                return links

            for link in links:
                if link["ifname"] == dev:
                    return link
            return None
        except Exception as e:
            logging.warning(f"{e}")

    def get_ip_addr_show(self, dev=None):
        """
        Retrieve the IP address information for a specified network device.

        This method executes the `ip -j addr show` command on the node to get
        the IP address information in JSON format for the specified device.
        If no device is specified, it defaults to the device name associated
        with the current instance.

        :param dev: The name of the network device (optional).
        :type dev: str, optional
        :return: The JSON output of the `ip -j addr show` command.
        :rtype: str
        :raises: Logs an error message if the command execution fails.
        """
        try:
            if not dev:
                dev = self.get_device_name()

            stdout, stderr = self.get_node().execute(
                f"ip -j addr show {dev}", quiet=True
            )
            return stdout
        except Exception as e:
            logging.error(
                f"Failed to get IP address show info for interface {self.get_name()}. Exception: {e}"
            )

    # fablib.Interface.get_ip_addr()
    def get_ip_addr_ssh(self, dev=None):
        """
        Gets the ip addr info for this interface.

        :return ip addr info
        :rtype: str
        """
        try:
            stdout, stderr = self.get_node().execute("ip -j addr list", quiet=True)

            addrs = json.loads(stdout)

            dev = self.get_device_name()
            # print(f"dev: {dev}")

            if dev is None:
                return addrs

            for addr in addrs:
                if addr["ifname"] == dev:
                    # Hack to make it backward compatible. Should return an object
                    return str(ipaddress.ip_address(addr["addr_info"][0]["local"]))

            return None
        except Exception as e:
            logging.warning(f"{e}")

    # fablib.Interface.get_ip_addr()
    def get_ips(self, family=None):
        """
        Gets a list of ips assigned to this interface.

        :return list of ips
        :rtype: list[str]
        """
        return_ips = []
        try:
            ip_addr = self.get_ip_addr()

            # print(f"{ip_addr}")

            for addr_info in ip_addr["addr_info"]:
                if family is None:
                    return_ips.append(addr_info["local"])
                else:
                    if addr_info["family"] == family:
                        return_ips.append(addr_info["local"])
        except Exception as e:
            logging.warning(f"{e}")

        return return_ips

    def get_fim(self):
        """
        Gets the node's FABRIC Information Model (fim) object. This method
        is used to access data at a lower level than FABlib.

        :return: the FABRIC model node
        :rtype: fim interface
        """
        return self.fim_interface

    def set_user_data(self, user_data: dict):
        """
        Set the user data for the interface.

        This method stores the given user data dictionary as a JSON string
        in the FIM object associated with the interface.

        :param user_data: The user data to be set.
        :type user_data: dict
        :return: None
        """
        self.get_fim().set_property(
            pname="user_data", pval=UserData(json.dumps(user_data))
        )

    def get_user_data(self):
        """
        Retrieve the user data for the interface.

        This method fetches the user data stored in the FIM object associated
        with the interface and returns it as a dictionary. If an error occurs,
        it returns an empty dictionary.

        :return: The user data dictionary.
        :rtype: dict
        """
        try:
            return json.loads(str(self.get_fim().get_property(pname="user_data")))
        except Exception as e:
            return {}

    def get_fablib_data(self):
        """
        Retrieve the 'fablib_data' from the user data.

        This method extracts and returns the 'fablib_data' field from the
        user data dictionary. If an error occurs or the field is not present,
        it returns an empty dictionary.

        :return: The 'fablib_data' dictionary.
        :rtype: dict
        """
        try:
            return self.get_user_data()["fablib_data"]
        except:
            return {}

    def set_fablib_data(self, fablib_data: dict):
        """
        Set the 'fablib_data' in the user data.

        This method updates the 'fablib_data' field in the user data dictionary
        and stores the updated user data back in the FIM.

        :param fablib_data: The 'fablib_data' to be set.
        :type fablib_data: dict
        :return: None
        """
        user_data = self.get_user_data()
        user_data["fablib_data"] = fablib_data
        self.set_user_data(user_data)

    def set_network(self, network: NetworkService):
        """
        Set the network for the interface.

        This method assigns the interface to the specified network. If the
        interface is already part of another network, it will be removed from
        the current network before being added to the new one.

        :param network: The network service to assign the interface to.
        :type network: NetworkService
        :return: The current instance with the updated network.
        :rtype: self
        """
        current_network = self.get_network()
        if current_network:
            current_network.remove_interface(self)

        network.add_interface(self)

        return self

    def set_ip_addr(self, addr: ipaddress = None, mode: str = None):
        """
        Set the IP address for the interface.

        This method assigns an IP address to the interface based on the provided
        address or allocation mode. If an address is provided, it will be allocated
        to the interface. If the mode is set to 'AUTO' and no address is provided,
        an IP address will be automatically allocated by the network.

        :param addr: The IP address to assign to the interface (optional).
        :type addr: ipaddress.IPv4Address or ipaddress.IPv6Address, optional
        :param mode: The mode for IP address allocation, e.g., `"auto"`, `"manual"`, or `"config"`.
        :type mode: str, optional
        :return: The current instance with the updated IP address.
        :rtype: self
        """
        fablib_data = self.get_fablib_data()
        if mode:
            fablib_data[self.MODE] = str(mode)

        mode = fablib_data[self.MODE]
        if addr:
            fablib_data[self.ADDR] = str(self.get_network().allocate_ip(addr))
        elif mode == self.AUTO:
            if self.get_network():
                fablib_data[self.ADDR] = str(self.get_network().allocate_ip())

        self.set_fablib_data(fablib_data)

        return self

    def get_ip_addr(self):
        """
        Retrieve the IP address assigned to the interface.

        This method returns the IP address assigned to the interface, either
        from the 'fablib_data' or by fetching it via SSH if not available in
        the stored data. If the MAC address is not available, it returns None.

        :return: The IP address assigned to the interface.
        :rtype: ipaddress.IPv4Address or ipaddress.IPv6Address or str or None
        """
        fablib_data = self.get_fablib_data()
        if self.ADDR in fablib_data:
            try:
                addr = ipaddress.ip_address(fablib_data[self.ADDR])
            except:
                addr = fablib_data[self.ADDR]
            return addr
        else:
            # get_ip_addr_ssh()
            if self.get_mac() is None:
                return None
            return self.get_ip_addr_ssh()

    def set_mode(self, mode: str = "config"):
        """
        Set the mode for the interface.

        This method sets the mode for the interface in the 'fablib_data'
        dictionary. The mode determines the configuration behavior of the
        interface.

        :param mode: The mode to set for the interface (default is "config"). Allowed values: `"auto"`, `"manual"`, or `"config"`..
        :type mode: str
        :return: The current instance with the updated mode.
        :rtype: self
        """
        fablib_data = self.get_fablib_data()
        fablib_data[self.MODE] = mode
        self.set_fablib_data(fablib_data)

        return self

    def get_mode(self):
        """
        Retrieve the mode of the interface.

        This method returns the current mode of the interface from the 'fablib_data'
        dictionary. If the mode is not set, it defaults to "config" and updates the
        'fablib_data' accordingly.

        :return: The mode of the interface.
        :rtype: str
        """
        fablib_data = self.get_fablib_data()
        if self.MODE not in fablib_data:
            self.set_mode(self.CONFIG)
            fablib_data = self.get_fablib_data()

        return fablib_data[self.MODE]

    def is_configured(self):
        """
        Check if the interface is configured.

        This method checks the 'fablib_data' dictionary to determine if the
        interface is marked as configured.

        :return: True if the interface is configured, False otherwise.
        :rtype: bool
        """
        fablib_data = self.get_fablib_data()
        if fablib_data:
            is_configured = fablib_data.get(self.CONFIGURED)
            if is_configured is None or not bool(is_configured):
                return False

        return True

    def config(self):
        """
        Configure the interface based on its mode and network settings. Called when a `.Node` is configured.

        This method configures the interface by setting its IP address and
        bringing it up. It checks the configuration mode and acts accordingly:
        - If the mode is 'AUTO' and no address is set, it automatically allocates an IP address.
        - If the mode is 'CONFIG' or 'AUTO', it configures the interface with the assigned IP address and subnet.
        - If the mode is 'MANUAL', it does not perform any automatic configuration.

        :return: None
        """
        network = self.get_network()
        if not network:
            logging.info(
                f"Interface {self.get_name()} not connected to a network, skipping configuration."
            )
            return

        fablib_data = self.get_fablib_data()
        addr = None
        if self.is_configured():
            addr = fablib_data.get(self.ADDR)
        else:
            fablib_data[self.CONFIGURED] = str(True)
            self.set_fablib_data(fablib_data)

        mode = fablib_data.get(self.MODE, self.MANUAL)

        if mode == self.AUTO and addr is None:
            fablib_data[self.ADDR] = str(self.get_network().allocate_ip())
            self.set_fablib_data(fablib_data)

        self.ip_link_up()

        if mode in [self.CONFIG, self.AUTO]:
            subnet = self.get_network().get_subnet()
            addr = fablib_data.get(self.ADDR)
            if addr and subnet:
                self.un_manage_interface()
                self.ip_link_up()
                self.ip_addr_add(addr=addr, subnet=ipaddress.ip_network(subnet))
        else:
            # Manual mode; do nothing.
            pass

    def add_mirror(self, port_name: str, name: str = "mirror", vlan: str = None):
        """
        Add Port Mirror Service

        :param port_name: Mirror Port Name
        :type port_name: String
        :param vlan: Mirror Port vlan
        :type vlan: String
        :param name: Name of the Port Mirror service
        :type name: String
        """
        self.get_slice().get_fim_topology().add_port_mirror_service(
            name=name,
            from_interface_name=port_name,
            from_interface_vlan=vlan,
            to_interface=self.get_fim(),
        )

    def delete(self):
        """
        Delete the interface by removing it from the corresponding network service
        """
        net = self.get_network()
        if net:
            net.remove_interface(self)
        if self.parent and self.parent.get_fim():
            self.parent.get_fim().remove_child_interface(name=self.get_name())

    def set_subnet(self, ipv4_subnet: str = None, ipv6_subnet: str = None):
        """
        Set subnet for the interface.
        Used only for interfaces connected to L3VPN service where each interface could be connected to multiple subnets

        :param ipv4_subnet: ipv4 subnet
        :type ipv4_subnet: str

        :param ipv6_subnet: ipv6 subnet
        :type ipv6_subnet: str

        :raises Exception in case invalid subnet string is specified.
        """
        try:
            labels = self.get_fim().labels
            if not labels:
                labels = Labels()
            if ipv4_subnet:
                ipaddress.ip_network(ipv4_subnet, strict=False)
                labels = Labels.update(labels, ipv4_subnet=ipv4_subnet)
            elif ipv6_subnet:
                ipaddress.ip_network(ipv6_subnet, strict=False)
                labels = Labels.update(labels, ipv6_subnet=ipv6_subnet)

            self.get_fim().set_property("labels", labels)
        except Exception as e:
            logging.error(f"Failed to set the ip subnet e: {e}")
            raise e

    def get_subnet(self):
        """
        Get Subnet associated with the interface

        :return: ipv4/ipv6 subnet associated with the interface
        :rtype: String
        """
        if self.get_fim() and self.get_fim().labels:
            if self.get_fim().labels.ipv4_subnet:
                return self.get_fim().labels.ipv4_subnet
            if self.get_fim().labels.ipv6_subnet:
                return self.get_fim().labels.ipv6_subnet

    def get_peer_subnet(self):
        """
        Get Peer Subnet associated with the interface

        :return: peer ipv4/ipv6 subnet associated with the interface
        :rtype: String
        """
        if self.get_fim() and self.get_fim().peer_labels:
            if self.get_fim().peer_labels.ipv4_subnet:
                return self.get_fim().peer_labels.ipv4_subnet
            if self.get_fim().peer_labels.ipv6_subnet:
                return self.get_fim().peer_labels.ipv6_subnet

    def get_peer_asn(self):
        """
        Get Peer ASN; Set only for Peered Interface using L3Peering via AL2S

        :return: peer asn
        :rtype: String
        """
        if self.get_fim() and self.get_fim().peer_labels:
            return self.get_fim().peer_labels.asn

    def get_peer_bgp_key(self):
        """
        Get Peer BGP Key; Set only for Peered Interface using L3Peering via AL2S

        :return: peer BGP Key
        :rtype: String
        """
        if self.get_fim() and self.get_fim().peer_labels:
            return self.get_fim().peer_labels.bgp_key

    def get_peer_account_id(self):
        """
        Get Peer Account Id associated with the interface

        :return: peer account id associated with the interface (Used when interface is peered to AWS via AL2S)
        :rtype: String
        """
        if self.get_fim() and self.get_fim().peer_labels:
            return self.get_fim().peer_labels.account_id

    def get_interfaces(self) -> List[Interface]:
        """
        Gets the interfaces attached to this fablib component's FABRIC component.

        :return: a list of the interfaces on this component.
        :rtype: List[Interface]
        """

        if not self.interfaces:
            self.interfaces = []
            for fim_interface in self.get_fim().interface_list:
                self.interfaces.append(
                    Interface(
                        component=self.get_component(),
                        fim_interface=fim_interface,
                        model=str(InterfaceType.SubInterface),
                        parent=self,
                    )
                )

        return self.interfaces

    def add_sub_interface(self, name: str, vlan: str, bw: int = 10):
        """
        Add a sub-interface to a dedicated NIC.

        This method adds a sub-interface to a NIC (Network Interface Card) with the specified
        name, VLAN (Virtual Local Area Network) ID, and bandwidth. It supports only specific
        NIC models.

        :param name: The name of the sub-interface.
        :type name: str

        :param vlan: The VLAN ID for the sub-interface.
        :type vlan: str

        :param bw: The bandwidth allocated to the sub-interface, in Gbps. Default is 10 Gbps.
        :type bw: int

        :raises Exception: If the NIC model does not support sub-interfaces.
        """
        if self.get_model() not in [
            Constants.CMP_NIC_ConnectX_5,
            Constants.CMP_NIC_ConnectX_6,
        ]:
            raise Exception(
                f"Sub interfaces are only supported for the following NIC models: "
                f"{Constants.CMP_NIC_ConnectX_5}, {Constants.CMP_NIC_ConnectX_6}"
            )

        # Hack for finding interfaces
        name = f"{self.get_name()}-{name}"

        if self.get_fim():
            child_interface = self.get_fim().add_child_interface(
                name=name, labels=Labels(vlan=vlan)
            )
            child_if_capacities = child_interface.get_property(pname="capacities")
            if not child_if_capacities:
                child_if_capacities = Capacities()
            child_if_capacities.bw = int(bw)
            child_interface.set_properties(capacities=child_if_capacities)
            if not self.interfaces:
                self.interfaces = []

            ch_iface = Interface(
                component=self.get_component(),
                fim_interface=child_interface,
                model=str(InterfaceType.SubInterface),
            )
            self.interfaces.append(ch_iface)
            return ch_iface

    def get_type(self) -> str:
        """
        Get Interface type
        :return: get interface type
        :rtype: String
        """
        if self.get_fim():
            return self.get_fim().type



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

"""
This module contains methods to work with FABRIC `facility ports`_.

.. _`facility ports`: https://learn.fabric-testbed.net/knowledge-base/glossary/#facility_port
"""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, List, Union

import jinja2
from fabrictestbed.slice_editor import Capacities, Labels
from tabulate import tabulate

from fabrictestbed_extensions.fablib.interface import Interface

if TYPE_CHECKING:
    from fim.user.node import Node as FimNode

    from fabrictestbed_extensions.fablib.slice import Slice


class FacilityPort:
    """
    A class for working with FABRIC facility ports.
    """

    fim_interface = None
    slice = None

    def __init__(self, slice: Slice, fim_interface: FimNode):
        """
        :param slice: the fablib slice to have this node on
        :type slice: Slice

        :param fim_interface:
        :type fim_interface: FimInterface
        """
        super().__init__()
        self.fim_interface = fim_interface
        self.slice = slice

    def __str__(self):
        """
        Creates a tabulated string describing the properties of the
        node.  Intended for printing node information.

        :return: Tabulated string of node information
        :rtype: String
        """
        table = [["name", self.get_name()]]

        return tabulate(table)

    def toJson(self):
        """
        Return a JSON representation of the facility port.
        """
        return json.dumps(self.toDict(), indent=4)

    def get_pretty_name_dict(self):
        """
        Return a mapping used when rendering table headers.
        """
        return {
            "name": "Name",
        }

    def toDict(self, skip=[]):
        """
        Return a Python `dict` representation of the facility port.
        """
        return {"name": str(self.get_name())}

    def get_template_context(self):
        return self.get_slice().get_template_context(self)

    def render_template(self, input_string):
        environment = jinja2.Environment()
        template = environment.from_string(input_string)
        output_string = template.render(self.get_template_context())

        return output_string

    def show(
        self, fields=None, output=None, quiet=False, colors=False, pretty_names=True
    ):
        """
        Get a human-readable representation of the facility port.
        """
        data = self.toDict()

        # fields = ["Name",
        #         ]

        if pretty_names:
            pretty_names_dict = self.get_pretty_name_dict()
        else:
            pretty_names_dict = {}

        table = self.get_fablib_manager().show_table(
            data,
            fields=fields,
            title="Facility Port",
            output=output,
            quiet=quiet,
            pretty_names_dict=pretty_names_dict,
        )

        return table

    def get_fim_interface(self) -> FimNode:
        """
        .. warning::
            Not recommended for most users.

        Gets the node's FABRIC Information Model (fim) object.  This
        method is used to access data at a lower level than FABlib.
        """
        return self.fim_interface

    def get_model(self) -> str:
        """
        Get fablib model name for the facility port.
        """
        return "Facility_Port"

    def get_name(self) -> str:
        """
        Gets the name of the FABRIC node.

        :return: the name of the node
        :rtype: String
        """
        return self.get_fim_interface().name

    def get_site(self) -> str:
        """
        Gets the site associated with the facility port.
        """
        return self.fim_interface.site

    @staticmethod
    def new_facility_port(
        slice: Slice = None,
        name: str = None,
        site: str = None,
        vlan: Union[List, str] = None,
        bandwidth: int = 10,
        mtu: int = None,
        labels: Labels = None,
        peer_labels: Labels = None,
    ) -> FacilityPort:
        """
        Create a new facility port in the given slice.

        You might want to :py:meth:`Slice.add_facility_port()`, in
        most cases.

        :param slice: The slice in which the facility port will be created.
        :param name: The name of the facility port.
        :param site: The site where the facility port will be located.
        :param vlan: A list or single string representing the VLANs for the facility port.
        :param bandwidth: The bandwidth capacity for the facility port, default is 10.
        :param mtu: MTU size
        :param labels: Labels associated with the facility port.
        :param peer_labels: Peer labels associated with the facility port.
        :return: A FacilityPort object representing the created facility port.
        """
        if not bandwidth:
            bandwidth = 10
        capacities = Capacities(bw=bandwidth)
        if mtu:
            capacities.mtu = mtu

        interfaces = None

        if vlan:
            index = 1
            interfaces = []
            if isinstance(vlan, str):
                vlan = [vlan]

            for v in vlan:
                iface_tuple = (
                    f"iface-{index}",
                    Labels(vlan=v),
                    capacities,
                )
                interfaces.append(iface_tuple)

        fim_facility_port = slice.get_fim_topology().add_facility(
            name=name,
            site=site,
            capacities=capacities,
            labels=labels,
            peer_labels=peer_labels,
            interfaces=interfaces,
        )
        return FacilityPort(slice, fim_facility_port)

    @staticmethod
    def get_facility_port(slice: Slice = None, facility_port: FimNode = None):
        """

        :param slice:
        :param facility_port:
        :return:
        """
        return FacilityPort(slice, facility_port)

    def get_slice(self) -> Slice:
        """
        Gets the fablib slice associated with this node.

        :return: the fablib slice on this node
        :rtype: Slice
        """
        return self.slice

    def get_interfaces(self) -> List[Interface]:
        """
        Gets a particular interface associated with a FABRIC node.

        Accepts either the interface name or a network_name.  If a
        network name is used this method will return the interface on
        the node that is connected to the network specified.  If a
        name and network_name are both used, the interface name will
        take precedence.

        :param name: interface name to search for
        :type name: str

        :param network_name: network name to search for

        :type name: str

        :raise Exception: if interface is not found

        :return: an interface on the node
        :rtype: Interface
        """
        ifaces = []
        for fim_interface in self.get_fim_interface().interface_list:
            ifaces.append(Interface(node=self, fim_interface=fim_interface))

        return ifaces



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Komal Thareja (kthare10@renci.org)
import json
import logging
import os
import re
from pathlib import Path
from typing import Dict, List, Union

import yaml
from atomicwrites import atomic_write

from fabrictestbed_extensions import __version__ as fablib_version
from fabrictestbed_extensions.fablib.constants import Constants
from fabrictestbed_extensions.utils.utils import Utils


class ConfigException(Exception):
    """
    An exception class to represent configuration errors.
    """

    pass


class Config:
    """
    A class that represents fablib configuration.
    """

    LOG_LEVELS = {
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
    }

    # Mapping of the required parameters and the corresponding Environment Variable name
    REQUIRED_ATTRS = {
        Constants.ORCHESTRATOR_HOST: {
            Constants.ENV_VAR: Constants.FABRIC_ORCHESTRATOR_HOST,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_ORCHESTRATOR_HOST,
        },
        Constants.CREDMGR_HOST: {
            Constants.ENV_VAR: Constants.FABRIC_CREDMGR_HOST,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_CREDMGR_HOST,
        },
        Constants.CORE_API_HOST: {
            Constants.ENV_VAR: Constants.FABRIC_CORE_API_HOST,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_CORE_API_HOST,
        },
        Constants.AM_HOST: {
            Constants.ENV_VAR: Constants.FABRIC_AM_HOST,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_AM_HOST,
        },
        Constants.TOKEN_LOCATION: {
            Constants.ENV_VAR: Constants.FABRIC_TOKEN_LOCATION,
            Constants.DEFAULT: Constants.DEFAULT_TOKEN_LOCATION,
        },
        Constants.PROJECT_ID: {
            Constants.ENV_VAR: Constants.FABRIC_PROJECT_ID,
        },
        Constants.BASTION_HOST: {
            Constants.ENV_VAR: Constants.FABRIC_BASTION_HOST,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_BASTION_HOST,
        },
        Constants.BASTION_USERNAME: {
            Constants.ENV_VAR: Constants.FABRIC_BASTION_USERNAME,
        },
        Constants.BASTION_KEY_LOCATION: {
            Constants.ENV_VAR: Constants.FABRIC_BASTION_KEY_LOCATION,
            Constants.DEFAULT: Constants.DEFAULT_BASTION_KEY_LOCATION,
        },
        Constants.FABLIB_VERSION: {Constants.DEFAULT: fablib_version},
        Constants.SLICE_PUBLIC_KEY_FILE: {
            Constants.ENV_VAR: Constants.FABRIC_SLICE_PUBLIC_KEY_FILE,
            Constants.DEFAULT: Constants.DEFAULT_SLICE_PUBLIC_KEY_FILE,
        },
        Constants.SLICE_PRIVATE_KEY_FILE: {
            Constants.ENV_VAR: Constants.FABRIC_SLICE_PRIVATE_KEY_FILE,
            Constants.DEFAULT: Constants.DEFAULT_SLICE_PRIVATE_KEY_FILE,
        },
        Constants.AVOID: {
            Constants.ENV_VAR: Constants.FABRIC_AVOID,
            Constants.DEFAULT: "",
        },
        Constants.SSH_COMMAND_LINE: {
            Constants.ENV_VAR: Constants.FABRIC_SSH_COMMAND_LINE,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_SSH_COMMAND_LINE,
        },
        Constants.LOG_LEVEL: {
            Constants.ENV_VAR: Constants.FABRIC_LOG_LEVEL,
            Constants.DEFAULT: Constants.DEFAULT_LOG_LEVEL,
        },
        Constants.LOG_FILE: {
            Constants.ENV_VAR: Constants.FABRIC_LOG_FILE,
            Constants.DEFAULT: Constants.DEFAULT_LOG_FILE,
        },
        Constants.BASTION_SSH_CONFIG_FILE: {
            Constants.ENV_VAR: Constants.FABRIC_BASTION_SSH_CONFIG_FILE,
            Constants.DEFAULT: Constants.DEFAULT_FABRIC_BASTION_SSH_CONFIG_FILE,
        },
    }

    REQUIRED_ATTRS_PRETTY_NAMES = {
        Constants.CREDMGR_HOST: "Credential Manager",
        Constants.ORCHESTRATOR_HOST: "Orchestrator",
        Constants.CORE_API_HOST: "Core API",
        Constants.AM_HOST: "Artifact Manager",
        Constants.TOKEN_LOCATION: "Token File",
        Constants.PROJECT_ID: "Project ID",
        Constants.PROJECT_NAME: "Project Name",
        Constants.BASTION_USERNAME: "Bastion Username",
        Constants.BASTION_KEY_LOCATION: "Bastion Private Key File",
        Constants.BASTION_HOST: "Bastion Host",
        Constants.BASTION_KEY_PASSPHRASE: "Bastion Private Key Passphrase",
        Constants.SLICE_PUBLIC_KEY_FILE: "Slice Public Key File",
        Constants.SLICE_PRIVATE_KEY_FILE: "Slice Private Key File",
        Constants.SLICE_PRIVATE_KEY_PASSPHRASE: "Slice Private Key Passphrase",
        Constants.LOG_FILE: "Log File",
        Constants.LOG_LEVEL: "Log Level",
        Constants.FABLIB_VERSION: "Version",
        Constants.AVOID: "Sites to avoid",
        Constants.DATA_DIR: "Data directory",
        Constants.SSH_COMMAND_LINE: "SSH Command Line",
        Constants.BASTION_SSH_CONFIG_FILE: "Bastion SSH Config File",
    }

    def __init__(
        self,
        fabric_rc: str = None,
        credmgr_host: str = None,
        orchestrator_host: str = None,
        core_api_host: str = None,
        am_host: str = None,
        token_location: str = None,
        project_id: str = None,
        bastion_username: str = None,
        bastion_key_location: str = None,
        log_level: str = Constants.DEFAULT_LOG_LEVEL,
        log_file: str = Constants.DEFAULT_LOG_FILE,
        data_dir: str = Constants.DEFAULT_DATA_DIR,
        offline: bool = True,
        **kwargs,
    ):
        """
        Constructor. Tries to get configuration from:

         - constructor parameters (high priority)
         - fabric_rc file (middle priority)
         - environment variables (low priority)
         - defaults (if needed and possible)

        """
        if fabric_rc is None:
            fabric_rc = Constants.DEFAULT_FABRIC_RC
            if os.path.exists(Constants.DEFAULT_FABRIC_CONFIG_DIR):
                Path(fabric_rc).touch()

        if fabric_rc and os.path.exists(fabric_rc):
            self.config_file_path = fabric_rc
        self.is_yaml = False
        self.runtime_config = {}
        self.offline = offline

        # Load from config file
        self.__load_configuration(file_path=fabric_rc, **kwargs)

        # Apply any parameters explicitly passed
        if credmgr_host is not None:
            self.set_credmgr_host(credmgr_host=credmgr_host)

        if orchestrator_host is not None:
            self.set_orchestrator_host(orchestrator_host=orchestrator_host)

        if core_api_host is not None:
            self.set_core_api_host(core_api_host=core_api_host)

        if am_host is not None:
            self.set_am_host(am_host=am_host)

        if token_location is not None:
            self.set_token_location(token_location=token_location)

        if project_id is not None:
            self.set_project_id(project_id=project_id)

        if bastion_username is not None:
            self.set_bastion_username(bastion_username=bastion_username)

        if bastion_key_location is not None:
            self.set_bastion_key_location(bastion_key_location=bastion_key_location)

        if log_level is not None:
            self.set_log_level(log_level=log_level)

        if log_file is not None:
            self.set_log_file(log_file=log_file)

        if data_dir is not None:
            self.set_data_dir(data_dir=data_dir)

        if self.get_ssh_command_line() is None:
            self.set_ssh_command_line(
                ssh_command_line=Constants.DEFAULT_FABRIC_SSH_COMMAND_LINE
            )

        self.required_check(partial=True)

        # Verify that Token file exists; any other checks cannot be done without this.
        token_location = self.get_token_location()
        if not os.path.exists(token_location):
            raise ConfigException(
                f"Token file does not exist, please provide the token at location: {token_location}!"
            )

    def __load_configuration(self, file_path, **kwargs):
        """
        Load the config parameters from config file;
        - Parses bashrc/yaml file and loads the parameters in dictionary

        :param file_path: path to the config file; can be bashrc or yaml
        :type file_path: str
        :param kwargs:
        :type kwargs:

        :raises ConfigException: if config file does not exist
        """
        if file_path and os.path.exists(file_path):
            if Utils.is_yaml_file(file_path=file_path):
                self.__load_yaml_file(file_path=file_path)
                self.is_yaml = True
            else:
                self.__load_fabric_rc_file(file_path=file_path)

        for k, v in kwargs.items():
            self.runtime_config[k] = v

    def __load_yaml_file(self, file_path):
        """
        Load the config from yml file
        :param file_path: path to the config file
        :type file_path: str

        :return True or False indicating success/failure for loading the config file
        :rtype bool
        """
        try:
            with open(file_path, "r") as file:
                config = yaml.safe_load(file)
                if (
                    isinstance(config, dict)
                    and config.get(Constants.RUNTIME_SECTION) is not None
                ):
                    for key, value in config.get(Constants.RUNTIME_SECTION).items():
                        if value is not None:
                            value = value.strip().strip("'").strip('"')
                            if key == Constants.AVOID:
                                value = value.split(",")
                            self.runtime_config[key] = value
                    self.runtime_config[Constants.FABLIB_VERSION] = fablib_version
                    return True
                return False
        except yaml.YAMLError:
            return False

    def __load_fabric_rc_file(self, file_path):
        """
        Load the config from bashrc file
        :param file_path: path to the config file
        :type file_path: str

        :return True or False indicating success/failure for loading the config file
        :rtype bool
        """
        ret_val = False
        with open(file_path, "r") as file:
            lines = file.readlines()

        export_pattern = re.compile(r"^export\s+([^=]+)=(.*)$", re.IGNORECASE)

        for line in lines:
            match = export_pattern.match(line.strip())
            if match:
                key, value = match.groups()
                value = value.strip()
                value = value.strip('"')
                value = value.strip("'")
                key = key.replace("FABRIC_", "").lower()
                if key == Constants.AVOID and value is not None and len(value):
                    value = value.split(",")
                self.runtime_config[key] = value
                ret_val = True
        self.runtime_config[Constants.FABLIB_VERSION] = fablib_version
        return ret_val

    def required_check(self, partial: bool = False):
        """
        Validates that all the required parameters are present in the loaded config
        Tries to load the parameters from environment variable or defauls if not available in config in that order

        :raises AttributeError if missing required parameters are found
        """
        errors = []

        for attr, attr_props in self.REQUIRED_ATTRS.items():
            if attr not in self.runtime_config or self.runtime_config.get(attr) is None:
                # Load from environment variables
                if (
                    attr_props.get(Constants.ENV_VAR)
                    and os.environ.get(attr_props.get(Constants.ENV_VAR)) is not None
                ):
                    self.runtime_config[attr] = os.environ.get(
                        attr_props.get(Constants.ENV_VAR)
                    )
                # Load defaults if available
                elif attr_props.get(Constants.DEFAULT) is not None:
                    self.runtime_config[attr] = attr_props.get(Constants.DEFAULT)
                # Ignore validation for this parameter if no default value exists and this partial check
                elif attr_props.get(Constants.DEFAULT) is None and partial:
                    continue
                else:
                    errors.append(f"{attr} is not set")

        if errors:
            logging.error(f"Failing Config: {self.runtime_config}")
            # TODO: define custom exception class to report errors,
            # and emit a more helpful error message with hints about
            # setting up environment variables or configuration file.
            raise AttributeError(
                f"Error initializing {self.__class__.__name__}: {errors}"
            )

    def get_config(self) -> Dict[str, str]:
        """
        Gets a dictionary mapping keywords to configured FABRIC environment
        variable values.

        :return: dictionary mapping keywords to FABRIC values
        :rtype: Dict[String, String]
        """
        return self.runtime_config

    def get_default_slice_public_key_file(self) -> str:
        """
        Gets the path to the default slice public key file.

        Important! Slice key management is underdevelopment and this
        functionality will likely change going forward.

        :return: the path to the slice public key on this fablib object
        :rtype: String
        """
        return self.runtime_config.get(Constants.SLICE_PUBLIC_KEY_FILE)

    def get_default_slice_private_key_file(self) -> str:
        """
        Gets the path to the default slice private key file.

        Important! Slices key management is underdevelopment and this
        functionality will likely change going forward.

        :return: the path to the slice private key on this fablib object
        :rtype: String
        """
        return self.runtime_config.get(Constants.SLICE_PRIVATE_KEY_FILE)

    def get_default_slice_private_key_passphrase(self) -> str:
        """
        Gets the passphrase to the default slice private key.

        Important! Slices key management is underdevelopment and this
        functionality will likely change going forward.

        :return: the passphrase to the slice private key on this fablib object
        :rtype: String
        """
        return self.runtime_config.get(Constants.SLICE_PRIVATE_KEY_PASSPHRASE)

    def get_credmgr_host(self) -> str:
        """
        Gets the credential manager host value.

        :return: the credential manager host
        :rtype: String
        """
        return self.get_config().get(Constants.CREDMGR_HOST)

    def set_credmgr_host(self, credmgr_host: str):
        """
        Sets the credential manager host value.
        :param credmgr_host: credential manager host site
        :type credmgr_host: String
        """
        self.runtime_config[Constants.CREDMGR_HOST] = credmgr_host

    def get_orchestrator_host(self) -> str:
        """
        Gets the orchestrator host value.

        :return: the orchestrator host site
        :rtype: String
        """
        return self.get_config().get(Constants.ORCHESTRATOR_HOST)

    def set_orchestrator_host(self, orchestrator_host: str):
        """
        Sets the Orchestrator host value.
        :param orchestrator_host: Orchestrator host
        :type orchestrator_host: String
        """
        self.runtime_config[Constants.ORCHESTRATOR_HOST] = orchestrator_host

    def get_core_api_host(self) -> str:
        """
        Gets the core_api host value.

        :return: the core_api host site
        :rtype: String
        """
        return self.get_config().get(Constants.CORE_API_HOST)

    def set_core_api_host(self, core_api_host: str):
        """
        Sets the core_api host value.
        :param core_api_host: core_api host
        :type core_api_host: String
        """
        self.runtime_config[Constants.CORE_API_HOST] = core_api_host

    def get_am_host(self) -> str:
        """
        Gets the am host value.

        :return: the am host site
        :rtype: String
        """
        return self.get_config().get(Constants.AM_HOST)

    def set_am_host(self, am_host: str):
        """
        Sets the am host value.
        :param am_host: core_api host
        :type am_host: String
        """
        self.runtime_config[Constants.AM_HOST] = am_host

    def get_token_location(self) -> str:
        """
        Gets the FABRIC token location.

        :return: FABRIC token location
        :rtype: String
        """
        return self.get_config().get(Constants.TOKEN_LOCATION)

    def set_token_location(self, token_location: str):
        """
        Sets the FABRIC token location.

        :param token_location: Token Location
        :type token_location: String
        """
        self.runtime_config[Constants.TOKEN_LOCATION] = token_location

    def get_bastion_username(self) -> str:
        """
        Gets the FABRIC Bastion username.

        :return: FABRIC Bastion username
        :rtype: String
        """
        return self.get_config().get(Constants.BASTION_USERNAME)

    def set_bastion_username(self, bastion_username: str):
        """
        Sets the FABRIC Bastion username.

        :param bastion_username: Bastion username
        :type bastion_username: String
        """
        self.runtime_config[Constants.BASTION_USERNAME] = bastion_username

    def get_bastion_key_location(self) -> str:
        """
        Gets the FABRIC Bastion key filename.

        :return: FABRIC Bastion key filename
        :rtype: String
        """
        return self.get_config().get(Constants.BASTION_KEY_LOCATION)

    def set_bastion_key_location(self, bastion_key_location: str):
        """
        Sets the FABRIC Bastion Key Location.

        :param bastion_key_location: Bastion Key Location
        :type bastion_key_location: String
        """
        self.runtime_config[Constants.BASTION_KEY_LOCATION] = bastion_key_location

    def get_bastion_key_passphrase(self) -> str:
        """
        Reads the FABRIC Bastion private key passphrase.

        :return: FABRIC Bastion key passphrase
        :rtype: String
        """
        return self.runtime_config.get(Constants.BASTION_KEY_PASSPHRASE)

    def get_bastion_key(self) -> Union[str, None]:
        """
        Reads the FABRIC Bastion private key file and returns the key.

        :return: FABRIC Bastion key string
        :rtype: String
        """
        if self.get_bastion_key_location() is None or not os.path.exists(
            self.get_bastion_key_location()
        ):
            return None
        return Utils.read_file_contents(file_path=self.get_bastion_key_location())

    def get_bastion_host(self) -> str:
        """
        Gets the FABRIC Bastion host address.

        :return: Bastion host public address
        :rtype: String
        """
        return self.get_config().get(Constants.BASTION_HOST)

    def get_project_id(self) -> str:
        """
        Get the Project Id:
        :return: Project Id
        :rtype: String
        """
        return self.get_config().get(Constants.PROJECT_ID)

    def set_project_id(self, project_id: str):
        """
        Set the Project Id:
        :param: project_id: Project Id
        :type: String
        """
        self.runtime_config[Constants.PROJECT_ID] = project_id

    def get_project_name(self):
        """
        Get the Project Name:
        :return: Project Name
        :rtype: String
        """
        return self.runtime_config.get(Constants.PROJECT_NAME)

    def set_log_level(self, log_level: str = "INFO"):
        """
        Sets the current log level for logging

        Options:  'DEBUG'
                  'INFO'
                  'WARNING'
                  'ERROR'
                  'CRITICAL'

        :param log_level: new log level
        :type str: Level
        """

        self.runtime_config[Constants.LOG_LEVEL] = log_level

    def get_log_level(self):
        """
        Get the current log level for logging

        :return log_level: new log level
        :rtype log_level: string
        """

        return self.runtime_config.get(Constants.LOG_LEVEL)

    def get_log_file(self) -> str:
        """
        Gets the current log file for logging

        :return log_file: new log level
        :rtype log_file: string
        """

        return self.runtime_config.get(Constants.LOG_FILE)

    def set_log_file(self, log_file: str):
        """
        Sets the current log file for logging

        :param log_file: new log level
        :type log_file: string
        """
        self.runtime_config[Constants.LOG_FILE] = log_file

    def get_data_dir(self) -> str:
        """
        Gets the data directory

        :return data_dir: data directory
        :rtype data_dir: string
        """

        return self.runtime_config.get(Constants.DATA_DIR)

    def set_data_dir(self, data_dir: str):
        """
        Sets the data directory

        :param data_dir: data directory
        :type data_dir: string
        """
        self.runtime_config[Constants.DATA_DIR] = data_dir

    def get_ssh_command_line(self):
        """
        Gets the data directory

        :return data_dir: data directory
        :rtype data_dir: string
        """
        return self.runtime_config.get(Constants.SSH_COMMAND_LINE)

    def set_ssh_command_line(self, ssh_command_line: str):
        """
        Sets the ssh command line

        :param ssh_command_line: ssh command line
        :type ssh_command_line: string
        """
        self.runtime_config[Constants.SSH_COMMAND_LINE] = ssh_command_line

    def get_bastion_ssh_config_file(self):
        """
        Gets the bastion ssh config file

        :return bastion_ssh_config_file: bastion_ssh_config_file
        :rtype bastion_ssh_config_file: string
        """
        return self.runtime_config.get(Constants.BASTION_SSH_CONFIG_FILE)

    def get_avoid(self):
        """
        Gets the avoid list

        :return avoid: avoid list
        :rtype avoid: string
        """
        return self.runtime_config.get(Constants.AVOID)

    def set_avoid(self, avoid: List[str]):
        """
        Sets the avoid list

        :param avoid: avoid list
        :type avoid: string
        """
        self.runtime_config[Constants.AVOID] = avoid

    def set_avoid_csv(self, avoid_csv: str = ""):
        """
        Sets the avoid site list from a CSV

        :param avoid_csv: avoid list
        :type avoid_csv: List of string
        """
        avoid_csv = avoid_csv.strip().strip('"').strip("'")

        avoid = []
        for site in avoid_csv.split(","):
            avoid.append(site.strip())

        self.set_avoid(avoid)

    def get_default_slice_private_key(self) -> Union[str, None]:
        """
        Gets the current default_slice_keys as a dictionary containg the
        public and private slice keys.

        Important! Slice key management is underdevelopment and this
        functionality will likely change going forward.

        :return: default_slice_key dictionary from superclass
        :rtype: Dict[String, String]
        """
        if self.get_default_slice_private_key_file() is not None and os.path.exists(
            self.get_default_slice_private_key_file()
        ):
            return Utils.read_file_contents(
                file_path=self.get_default_slice_private_key_file()
            )
        return None

    def get_default_slice_public_key(self) -> Union[str, None]:
        """
        Gets the current default_slice_keys as a dictionary containg the
        public and private slice keys.

        Important! Slice key management is underdevelopment and this
        functionality will likely change going forward.

        :return: default_slice_key dictionary from superclass
        :rtype: Dict[String, String]
        """
        if self.get_default_slice_public_key_file() is not None and os.path.exists(
            self.get_default_slice_public_key_file()
        ):
            return Utils.read_file_contents(
                file_path=self.get_default_slice_public_key_file()
            )
        return None

    @staticmethod
    def get_image_names() -> List[str]:
        """
        Gets a list of available image names.

        This is statically defined for now. Eventually, images will be managed dynamically.

        :return: list of image names as strings
        :rtype: list[str]
        """
        return Constants.IMAGE_NAMES

    def get_config_pretty_names_dict(self) -> Dict[str, str]:
        """
        Return PRETTY Names for the config

        :return: Dict of Pretty Names
        :rtype: Dict[str, str]
        """
        return self.REQUIRED_ATTRS_PRETTY_NAMES

    def save_config(self):
        """
        Write the configuration file.
        """
        if self.config_file_path is None:
            print("Config file path not set!")
            return

        if self.is_yaml:
            # Write the dictionary to the YAML file
            with atomic_write(self.config_file_path, overwrite=True) as f:
                yaml.dump(self.runtime_config, f, default_flow_style=False)
        else:
            with atomic_write(self.config_file_path, overwrite=True) as f:
                for attr, attr_props in self.REQUIRED_ATTRS.items():
                    env_var = attr_props.get(Constants.ENV_VAR)
                    value = self.runtime_config.get(attr)
                    if value is None:
                        value = ""
                    if env_var:
                        f.write(f"export {env_var}={value}\n")

    def setup_logging(self):
        """
        Create log file if it doesn't exist; setup logger
        """
        try:
            for handler in logging.root.handlers[:]:
                logging.root.removeHandler(handler)
        except Exception as e:
            print(f"Exception from removeHandler: {e}")
            pass

        try:
            if self.get_log_file() and not os.path.isdir(
                os.path.dirname(self.get_log_file())
            ):
                os.makedirs(os.path.dirname(self.get_log_file()))
        except Exception:
            logging.warning(
                f"Failed to create log_file directory: {os.path.dirname(self.get_log_file())}"
            )

        if self.get_log_file() and self.get_log_level():
            logging.basicConfig(
                filename=self.get_log_file(),
                level=self.LOG_LEVELS[self.get_log_level()],
                format="[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s",
                datefmt="%H:%M:%S",
            )


if __name__ == "__main__":
    config1 = Config(fabric_rc="./fabric_rc", credmgr_host="abc.def.com")
    assert config1.get_credmgr_host() == "abc.def.com"
    print(json.dumps(config1.runtime_config, indent=4))

    config2 = Config(fabric_rc="./fabric_rc.yml", credmgr_host="abc.def.com")
    assert config2.get_credmgr_host() == "abc.def.com"
    print(json.dumps(config2.runtime_config, indent=4))



import unittest

from fabrictestbed_extensions.fablib.fablib import FablibManager


class FabNet4ExtTest(unittest.TestCase):
    """
    Run some basic tests against the testbed.
    """

    fablib = None
    slice_name = "MySlice-fabnet-v4ext"
    node1_name = "Node1"
    node2_name = "Node2"

    network1_name = "net1"
    network2_name = "net2"

    node1_nic_name = "nic1"
    node2_nic_name = "nic2"

    def setUp(self) -> None:
        self.fablib = FablibManager()
        self.fablib.show_config()

    def test_1_fabnetv4_ext_create_slice(self):
        """
        Create a slice with a single node, and echo a message from the node.
        """
        [site1, site2] = self.fablib.get_random_sites(count=2)
        print(f"Sites: {site1}, {site2}")

        # Create Slice
        slice = self.fablib.new_slice(name=self.slice_name)

        # Node1
        node1 = slice.add_node(name=self.node1_name, site=site1)
        iface1 = node1.add_component(
            model="NIC_Basic", name=self.node1_nic_name
        ).get_interfaces()[0]

        # Node2
        node2 = slice.add_node(name=self.node2_name, site=site2)
        iface2 = node2.add_component(
            model="NIC_Basic", name=self.node2_nic_name
        ).get_interfaces()[0]

        # NetworkS
        net1 = slice.add_l3network(
            name=self.network1_name, interfaces=[iface1], type="IPv4Ext"
        )
        net2 = slice.add_l3network(
            name=self.network2_name, interfaces=[iface2], type="IPv4Ext"
        )

        # Submit Slice Request
        slice.submit()

    def test_2_fabnetv4_ext_make_ip_routable(self):
        slice = self.fablib.get_slice(name=self.slice_name)
        network1 = slice.get_network(name=self.network1_name)
        network1_available_ips = network1.get_available_ips()
        network1.show()

        network2 = slice.get_network(name=self.network2_name)
        network2_available_ips = network2.get_available_ips()
        network2.show()

        try:
            # Enable Public IPv6 make_ip_publicly_routable
            network1.make_ip_publicly_routable(ipv4=[str(network1_available_ips[0])])

            # Enable Public IPv6 make_ip_publicly_routable
            network2.make_ip_publicly_routable(ipv4=[str(network2_available_ips[0])])

            slice.submit()

        except Exception as e:
            print(f"Exception: {e}")
            import traceback

            traceback.print_exc()

    def test_3_fabnetv4_ext_connectivity(self):
        slice = self.fablib.get_slice(name=self.slice_name)
        network1 = slice.get_network(name=self.network1_name)
        network2 = slice.get_network(name=self.network2_name)

        node1 = slice.get_node(name=self.node1_name)
        node1_iface = node1.get_interface(network_name=self.network1_name)
        node1_addr = network1.get_public_ips()[0]
        node1_iface.ip_addr_add(addr=node1_addr, subnet=network1.get_subnet())

        node1.ip_route_add(subnet=network2.get_subnet(), gateway=network1.get_gateway())

        # Add route to external network
        stdout, stderr = node1.execute(
            f"sudo ip route add 0.0.0.0/1 via {network1.get_gateway()}"
        )

        stdout, stderr = node1.execute(f"ip addr show {node1_iface.get_device_name()}")
        stdout, stderr = node1.execute(f"ip route list")

        node2 = slice.get_node(name=self.node2_name)
        node2_iface = node2.get_interface(network_name=self.network2_name)
        node2_addr = network2.get_public_ips()[0]
        node2_iface.ip_addr_add(addr=node2_addr, subnet=network2.get_subnet())

        node2.ip_route_add(subnet=network1.get_subnet(), gateway=network2.get_gateway())

        # Add route to external network
        stdout, stderr = node2.execute(
            f"sudo ip route add 0.0.0.0/1 via {network2.get_gateway()}"
        )

        stdout, stderr = node2.execute(f"ip addr show {node2_iface.get_device_name()}")
        stdout, stderr = node2.execute(f"ip route list")

        node1 = slice.get_node(name=self.node1_name)
        node2 = slice.get_node(name=self.node2_name)

        node2_addr = node2.get_interface(network_name=self.network2_name).get_ip_addr()

        stdout, stderr = node1.execute(f"ping -c 5 {node2_addr}")

        # Verify external connectivity
        stdout, stderr = node1.execute(
            f"sudo ping -c 5 -I {node1_iface.get_device_name()} bing.com"
        )

        stdout, stderr = node1.execute(
            f"sudo ping -c 5 -I {node2_iface.get_device_name()} bing.com"
        )

    def test_3_fabnetv4_ext_delete(self):
        slice = self.fablib.get_slice(name=self.slice_name)
        slice.delete()


if __name__ == "__main__":
    unittest.main()
